# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class SurvivalAnalysisMixin(object):

    def score(self, X, y):
        from .metrics import concordance_index_censored
        name_event, name_time = y.dtype.names

        result = concordance_index_censored(y[name_event], y[name_time], self.predict(X))
        return result[0]



from os.path import join, dirname

import numpy
from numpy.testing import TestCase, run_module_suite, assert_array_almost_equal
from sklearn.base import BaseEstimator
from sklearn.datasets import load_iris
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.svm import SVC

from sksurv.column import categorical_to_numeric
from sksurv.datasets import load_whas500
from sksurv.meta import Stacking, MeanEstimator
from sksurv.metrics import concordance_index_censored
from sksurv.svm import FastSurvivalSVM
from sksurv.linear_model import CoxPHSurvivalAnalysis


class _NoFitEstimator(BaseEstimator):
    pass


class _NoPredictDummy(BaseEstimator):
    def fit(self, X, y):
        pass


class _PredictDummy(BaseEstimator):
    def fit(self, X, y):
        pass

    def predict(self, X):
        pass


class _PredictProbaDummy(BaseEstimator):
    def fit(self, X, y):
        pass

    def predict_proba(self, X):
        pass


class TestStackingClassifier(TestCase):
    def test_base_no_fit(self):
        self.assertRaisesRegex(TypeError,
                               "All base estimators should implement fit and predict/predict_proba (.+) doesn't",
                               Stacking, _PredictDummy, [('m1', _NoFitEstimator)])

    def test_base_no_predict(self):
        self.assertRaisesRegex(TypeError,
                               "All base estimators should implement fit and predict/predict_proba (.+) doesn't",
                               Stacking, _PredictDummy, [('m1', _NoPredictDummy)])

    def test_meta_no_fit(self):
        self.assertRaisesRegex(TypeError,
                               "meta estimator should implement fit (.+) doesn't",
                               Stacking, _NoFitEstimator, [('m1', _PredictDummy)])

    def test_names_not_unique(self):
        self.assertRaisesRegex(ValueError,
                               "Names provided are not unique: \('m1', 'm2', 'm1'\)",
                               Stacking, _NoFitEstimator,
                               [('m1', _PredictDummy), ('m2', _PredictDummy), ('m1', _PredictDummy)])

    def test_fit(self):
        data = load_iris()
        x = data["data"]
        y = data["target"]

        meta = Stacking(LogisticRegression(), [('tree', DecisionTreeClassifier(max_depth=1, random_state=0)),
                                               ('svm', SVC(probability=True, random_state=0))])
        self.assertEqual(2, len(meta))
        meta.fit(x, y)

        p = meta._predict_estimators(x)
        self.assertTupleEqual((x.shape[0], 3 * 2), p.shape)

        self.assertTupleEqual((3, 3 * 2), meta.meta_estimator.coef_.shape)

    def test_fit_sample_weights(self):
        data = load_iris()
        x = data["data"]
        y = data["target"]

        meta = Stacking(LogisticRegression(), [('tree', DecisionTreeClassifier(max_depth=1, random_state=0)),
                                               ('svm', SVC(probability=True, random_state=0))])

        sample_weight = numpy.random.RandomState(0).uniform(size=x.shape[0])
        meta.fit(x, y, tree__sample_weight=sample_weight, svm__sample_weight=sample_weight)

    def test_set_params(self):
        meta = Stacking(LogisticRegression(), [('tree', DecisionTreeClassifier(max_depth=1, random_state=0)),
                                               ('svm', SVC(probability=True, random_state=0))],
                        probabilities=True)
        self.assertEqual(2, len(meta))
        meta.set_params(tree__min_samples_split=7, svm__C=0.05)

        self.assertEqual(7, meta.get_params()["tree__min_samples_split"])
        self.assertEqual(0.05, meta.get_params()["svm__C"])
        self.assertIsInstance(meta.get_params()["meta_estimator"], LogisticRegression)
        self.assertTrue(meta.get_params()["probabilities"])

        meta.set_params(meta_estimator=DecisionTreeClassifier(), probabilities=False)
        self.assertIsInstance(meta.get_params()["meta_estimator"], DecisionTreeClassifier)
        self.assertFalse(meta.get_params()["probabilities"])

    def test_predict(self):
        data = load_iris()
        x = data["data"]
        y = data["target"]

        meta = Stacking(LogisticRegression(multi_class='multinomial', solver='lbfgs'),
                        [('tree', DecisionTreeClassifier(max_depth=1, random_state=0)),
                         ('svm', SVC(probability=True, random_state=0))])
        self.assertEqual(2, len(meta))
        meta.fit(x, y)
        p = meta.predict(x)
        acc = accuracy_score(y, p)

        self.assertGreaterEqual(acc, 0.98)

    def test_predict_proba(self):
        data = load_iris()
        x = data["data"]
        y = data["target"]

        meta = Stacking(LogisticRegression(multi_class='multinomial', solver='lbfgs'),
                        [('tree', DecisionTreeClassifier(max_depth=1, random_state=0)),
                         ('svm', SVC(probability=True, random_state=0))])
        meta.fit(x, y)
        p = meta.predict_proba(x)

        scores = numpy.empty(3)
        for i, c in enumerate(meta.meta_estimator.classes_):
            scores[i] = roc_auc_score(numpy.asarray(y == c, dtype=int), p[:, i])

        assert_array_almost_equal(numpy.array([1.0, 0.9986, 0.9986]), scores)

    def test_predict_log_proba(self):
        data = load_iris()
        x = data["data"]
        y = data["target"]

        meta = Stacking(LogisticRegression(multi_class='multinomial', solver='lbfgs'),
                        [('tree', DecisionTreeClassifier(max_depth=1, random_state=0)),
                         ('svm', SVC(probability=True, random_state=0))])
        meta.fit(x, y)
        p = meta.predict_log_proba(x)

        scores = numpy.empty(3)
        for i, c in enumerate(meta.meta_estimator.classes_):
            scores[i] = roc_auc_score(numpy.asarray(y == c, dtype=int), p[:, i])

        assert_array_almost_equal(numpy.array([1.0, 0.9986, 0.9986]), scores)


class TestStackingSurvivalAnalysis(TestCase):
    def setUp(self):
        x, self.y = load_whas500()
        self.x = categorical_to_numeric(x)

    def test_fit(self):
        meta = Stacking(MeanEstimator(),
                        [('coxph', CoxPHSurvivalAnalysis()),
                         ('svm', FastSurvivalSVM(random_state=0))],
                        probabilities=False)
        self.assertEqual(2, len(meta))
        meta.fit(self.x.values, self.y)

        p = meta._predict_estimators(self.x.values)
        self.assertTupleEqual((self.x.shape[0], 2), p.shape)

    def test_set_params(self):
        meta = Stacking(_PredictDummy(),
                        [('coxph', CoxPHSurvivalAnalysis()),
                         ('svm', FastSurvivalSVM(random_state=0))],
                        probabilities=False)

        meta.set_params(coxph__alpha=1.0, svm__alpha=0.4132)

        self.assertEqual(1.0, meta.get_params()["coxph__alpha"])
        self.assertEqual(0.4132, meta.get_params()["svm__alpha"])

    def test_predict(self):
        meta = Stacking(MeanEstimator(),
                        [('coxph', CoxPHSurvivalAnalysis()),
                         ('svm', FastSurvivalSVM(random_state=0))],
                        probabilities=False)

        meta.fit(self.x.values, self.y)

        # result is different if randomForestSRC has not been compiled with OpenMP support
        p = meta.predict(self.x.values)
        actual_cindex = concordance_index_censored(self.y['fstat'], self.y['lenfol'], p)

        expected_cindex = numpy.array([0.7848807, 58983, 16166, 0, 119])
        assert_array_almost_equal(expected_cindex, actual_cindex)

    def test_predict_proba(self):
        meta = Stacking(_PredictDummy(),
                        [('coxph', CoxPHSurvivalAnalysis()),
                         ('svm', FastSurvivalSVM(random_state=0))],
                        probabilities=False)

        self.assertRaisesRegex(AttributeError,
                               "'_PredictDummy' object has no attribute 'predict_proba'",
                               getattr, meta, "predict_proba")


if __name__ == '__main__':
    run_module_suite()



from numpy.testing import TestCase, run_module_suite, assert_array_almost_equal

import numpy
import pandas
from sklearn.base import clone
from sklearn.metrics.pairwise import pairwise_kernels

from sksurv.kernels import clinical_kernel, ClinicalKernelTransform


def _get_expected_matrix(with_ordinal=True, with_nominal=True, with_continuous=True):
    mat_age = numpy.array([[1., 0.9625, 0.925, 0.575, 0.],
                           [0.9625, 1., 0.9625, 0.6125, 0.0375],
                           [0.925, 0.9625, 1., 0.6500, 0.075],
                           [0.575, 0.6125, 0.6500, 1., 0.425],
                           [0., 0.0375, 0.075, 0.425, 1.]])

    mat_node_size = numpy.array([[1., 2/3, 2/3, 1/3, 2/3],
                                [2/3, 1., 1/3, 0., 1.],
                                [2/3, 1/3, 1., 2/3, 1/3],
                                [1/3, 0., 2/3, 1., 0.],
                                [2/3, 1., 1/3, 0., 1.]])

    mat_node_spread = numpy.array([[1., 0., 1., 0.5, 0.],
                                   [0., 1., 0., 0.5, 1.],
                                   [1., 0., 1., 0.5, 0.],
                                   [0.5, 0.5, 0.5, 1., 0.5],
                                   [0., 1., 0., 0.5, 1.]])

    mat_metastasis = numpy.array([[1, 0, 1, 1, 0],
                                  [0, 1, 0, 0, 1],
                                  [1, 0, 1, 1, 0],
                                  [1, 0, 1, 1, 0],
                                  [0, 1, 0, 0, 1]], dtype=float)

    included = []
    if with_continuous:
        included.append(mat_age)
    if with_ordinal:
        included.append(mat_node_size)
        included.append(mat_node_spread)
    if with_nominal:
        included.append(mat_metastasis)

    expected = included[0]
    for i in range(1, len(included)):
        expected += included[i]
    expected /= len(included)

    return expected


class TestClinicalKernel(TestCase):
    def setUp(self):
        data = pandas.DataFrame({'age': [20, 23, 26, 54, 100],
                                 'lymph node size': [2, 1, 3, 4, 1],
                                 'lymph node spread': ['distant', 'none', 'distant', 'close', 'none'],
                                 'metastasis': ['yes', 'no', 'yes', 'yes', 'no']})
        data['lymph node size'] = pandas.Categorical(data['lymph node size'],
                                                     categories=[1, 2, 3, 4],
                                                     ordered=True)
        data['lymph node spread'] = pandas.Categorical(data['lymph node spread'],
                                                       categories=['none', 'close', 'distant'],
                                                       ordered=True)
        data['metastasis'] = pandas.Categorical(data['metastasis'],
                                                categories=['no', 'yes'],
                                                ordered=False)
        self.data = data

    def test_clinical_kernel_1(self):
        mat = clinical_kernel(self.data)
        expected = _get_expected_matrix()

        assert_array_almost_equal(expected, mat, 4)

    def test_clinical_kernel_no_ordinal(self):
        mat = clinical_kernel(self.data.drop(['lymph node size', 'lymph node spread'], axis=1))
        expected = _get_expected_matrix(with_ordinal=False)
        assert_array_almost_equal(expected, mat, 4)

    def test_clinical_kernel_no_nominal(self):
        mat = clinical_kernel(self.data.drop('metastasis', axis=1))
        expected = _get_expected_matrix(with_nominal=False)
        assert_array_almost_equal(expected, mat, 4)

    def test_clinical_kernel_no_continuous(self):
        mat = clinical_kernel(self.data.drop('age', axis=1))
        expected = _get_expected_matrix(with_continuous=False)
        assert_array_almost_equal(expected, mat, 4)

    def test_clinical_kernel_only_nominal(self):
        mat = clinical_kernel(self.data.drop(['age', 'lymph node size', 'lymph node spread'], axis=1))
        expected = _get_expected_matrix(with_continuous=False, with_ordinal=False)
        assert_array_almost_equal(expected, mat, 4)

    def test_clinical_kernel_x_and_y(self):
        mat = clinical_kernel(self.data.iloc[:3, :], self.data.iloc[3:, :])
        m = _get_expected_matrix()
        expected = m[:3:, 3:]

        assert_array_almost_equal(expected, mat, 4)

    def test_fit_error_ndim(self):
        t = ClinicalKernelTransform()

        self.assertRaisesRegex(ValueError, "expected 2d array, but got 1",
                               t.fit, numpy.random.randn(31))

        self.assertRaisesRegex(ValueError, "expected 2d array, but got 3",
                               t.fit, numpy.random.randn(31, 20, 2))

    def test_kernel_transform(self):
        t = ClinicalKernelTransform()

        t.fit(self.data)
        mat = t.transform(t.X_fit_)

        expected = _get_expected_matrix()

        assert_array_almost_equal(expected, mat, 4)

    def test_kernel_transform_x_and_y(self):
        t = ClinicalKernelTransform(fit_once=True)
        t.prepare(self.data)
        x_num = t.X_fit_.copy()

        t.fit(x_num[:3, :])
        mat = t.transform(x_num[3:, :])

        m = _get_expected_matrix()
        expected = m[:3, 3:].T

        assert_array_almost_equal(expected, mat, 4)

    def test_kernel_transform_feature_mismatch(self):
        t = ClinicalKernelTransform()
        t.fit(self.data)

        self.assertRaisesRegex(ValueError, 'expected array with 4 features, but got 17',
                               t.transform, numpy.zeros((2, 17), dtype=float))

    def test_pairwise(self):
        t = ClinicalKernelTransform()
        t.fit(self.data)

        mat = pairwise_kernels(t.X_fit_, t.X_fit_,
                               metric=t.pairwise_kernel, n_jobs=1)

        expected = _get_expected_matrix()

        assert_array_almost_equal(expected, mat, 4)

    def test_pairwise_x_and_y(self):
        t = ClinicalKernelTransform()
        t.fit(self.data)

        mat = pairwise_kernels(t.X_fit_[:3, :], t.X_fit_[3:, :],
                               metric=t.pairwise_kernel, n_jobs=1)

        m = _get_expected_matrix()
        expected = m[:3:, 3:]

        assert_array_almost_equal(expected, mat, 4)

    def test_pairwise_x_and_y_error_shape(self):
        t = ClinicalKernelTransform()
        t.fit(self.data)

        self.assertRaisesRegex(ValueError, "X and Y have different number of features",
                               t.pairwise_kernel, self.data.iloc[0, :], self.data.iloc[1, :2])

    def test_pairwise_no_nominal(self):
        t = ClinicalKernelTransform()
        t.fit(self.data.drop('metastasis', axis=1))

        mat = pairwise_kernels(t.X_fit_[:3, :], t.X_fit_[3:, :],
                               metric=t.pairwise_kernel, n_jobs=1)

        expected = _get_expected_matrix(with_nominal=False)
        assert_array_almost_equal(expected[:3:, 3:], mat, 4)

    def test_call_function(self):
        t = ClinicalKernelTransform(fit_once=True)
        t.prepare(self.data)

        mat = t(t.X_fit_, t.X_fit_)
        expected = _get_expected_matrix()
        assert_array_almost_equal(expected, mat, 4)

    def test_call_function_x_and_y(self):
        t = ClinicalKernelTransform(fit_once=True)
        t.prepare(self.data)

        mat = t(t.X_fit_[:3, :], t.X_fit_[3:, :])
        m = _get_expected_matrix()
        expected = m[:3:, 3:]

        assert_array_almost_equal(expected, mat, 4)

    def test_pairwise_feature_mismatch(self):
        t = ClinicalKernelTransform()
        t.fit(self.data)

        self.assertRaisesRegex(ValueError, 'Incompatible dimension for X and Y matrices: '
                                           'X.shape\[1\] == 4 while Y.shape\[1\] == 17',
                               pairwise_kernels, t.X_fit_, numpy.zeros((2, 17), dtype=float),
                               metric=t.pairwise_kernel, n_jobs=1)

    def test_prepare(self):
        t = ClinicalKernelTransform(fit_once=True)
        t.prepare(self.data)

        copy = clone(t).fit(t.X_fit_)
        mat = copy.transform(t.X_fit_[:4, :])

        expected = _get_expected_matrix()
        assert_array_almost_equal(expected[:4, :], mat, 4)

    def test_prepare_error_fit_once(self):
        t = ClinicalKernelTransform(fit_once=False)

        self.assertRaisesRegex(ValueError, "prepare can only be used if fit_once parameter is set to True",
                               t.prepare, self.data)

    def test_prepare_error_type(self):
        t = ClinicalKernelTransform(fit_once=True)

        self.assertRaisesRegex(TypeError, 'X must be a pandas DataFrame',
                               t.prepare, [[0, 1], [1, 2], [4, 3], [6, 5]])

    def test_prepare_error_dtype(self):
        t = ClinicalKernelTransform(fit_once=True)
        data = pandas.DataFrame({"age": [12, 61, 18, 21, 57, 17],
                                 "date": numpy.array(
                                     ["2016-01-01", "1954-06-30", "1999-03-01", "2005-02-25", "2112-12-31",
                                      "1731-09-16"], dtype='datetime64')})

        self.assertRaisesRegex(TypeError, 'unsupported dtype: dtype\(.+\)',
                               t.prepare, data)

    def test_feature_mismatch(self):
        x = self.data.iloc[:, :2]
        y = self.data.iloc[:, 2:]
        self.assertRaisesRegex(ValueError, 'columns do not match',
                               clinical_kernel, x, y)

        y = numpy.zeros((10, 17))
        self.assertRaisesRegex(ValueError, 'x and y have different number of features',
                               clinical_kernel, x, y)


if __name__ == '__main__':
    run_module_suite()



import unittest
from functools import partial
from os.path import join, dirname

from nose.plugins.attrib import attr
import numpy
from numpy.testing import TestCase, run_module_suite, assert_array_almost_equal, assert_array_equal
from sklearn.decomposition import KernelPCA
from sklearn.metrics import mean_squared_error

from sksurv.bintrees import AVLTree, RBTree
from sksurv.column import encode_categorical, standardize
from sksurv.datasets import load_whas500, get_x_y
from sksurv.io import loadarff
from sksurv.kernels import ClinicalKernelTransform
from sksurv.metrics import concordance_index_censored
from sksurv.svm._prsvm import survival_constraints_simple
from sksurv.svm.naive_survival_svm import NaiveSurvivalSVM
from sksurv.svm.survival_svm import FastSurvivalSVM, FastKernelSurvivalSVM, SurvivalCounter, \
    OrderStatisticTreeSurvivalCounter

WHAS500_NOTIES_FILE = join(dirname(__file__), 'data', 'whas500-noties.arff')


class TestSurvivalSVM(TestCase):
    def test_alpha_negative(self):
        x = numpy.zeros((100, 10))
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=100)
        y['event'] = numpy.ones(100, dtype=bool)
        y['time'] = numpy.arange(100, dtype=float)

        ssvm = FastSurvivalSVM(alpha=-1)
        self.assertRaisesRegex(ValueError, "alpha must be positive",
                               ssvm.fit, x, y)

    def test_rank_ratio_out_of_bounds(self):
        x = numpy.zeros((100, 10))
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=100)
        y['event'] = numpy.ones(100, dtype=bool)
        y['time'] = numpy.arange(100, dtype=float)

        ssvm = FastSurvivalSVM(rank_ratio=-1)
        self.assertRaisesRegex(ValueError, "rank_ratio must be in \[0; 1\]",
                               ssvm.fit, x, y)

        ssvm.set_params(rank_ratio=1.2)
        self.assertRaisesRegex(ValueError, "rank_ratio must be in \[0; 1\]",
                               ssvm.fit, x, y)

        ssvm.set_params(rank_ratio=numpy.nan)
        self.assertRaisesRegex(ValueError, "rank_ratio must be in \[0; 1\]",
                               ssvm.fit, x, y)

        ssvm.set_params(rank_ratio=numpy.inf)
        self.assertRaisesRegex(ValueError, "rank_ratio must be in \[0; 1\]",
                               ssvm.fit, x, y)

    def test_regression_not_supported(self):
        x = numpy.zeros((100, 10))
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=100)
        y['event'] = numpy.ones(100, dtype=bool)
        y['time'] = numpy.arange(100, dtype=float)

        ssvm = FastSurvivalSVM(rank_ratio=0, optimizer='simple')
        self.assertRaisesRegex(ValueError,
                               "optimizer 'simple' does not implement regression objective",
                               ssvm.fit, x, y)

        ssvm.set_params(optimizer='PRSVM')
        self.assertRaisesRegex(ValueError,
                               "optimizer 'PRSVM' does not implement regression objective",
                               ssvm.fit, x, y)

    def test_unknown_optimizer(self):
        x = numpy.zeros((100, 10))
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=100)
        y['event'] = numpy.ones(100, dtype=bool)
        y['time'] = numpy.arange(1, 101, dtype=float)

        ssvm = FastSurvivalSVM(rank_ratio=0, optimizer='random stuff')
        self.assertRaisesRegex(ValueError,
                               "unknown optimizer: random stuff",
                               ssvm.fit, x, y)

    def test_y_no_array(self):
        x = numpy.zeros((100, 10))
        y = [numpy.ones(100, dtype=bool), numpy.arange(100)]

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               'y must be a structured array with the first field'
                               ' being a binary class event indicator and the second field'
                               ' the time of the event/censoring',
                               rsvm.fit, x, y)

    def test_only_one_label(self):
        x = numpy.zeros((100, 10))
        y = numpy.ones(100, dtype=int)

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               'y must be a structured array with the first field'
                               ' being a binary class event indicator and the second field'
                               ' the time of the event/censoring',
                               rsvm.fit, x, y)

    def test_y_one_field(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.ones(dtype=[('event', bool)], shape=10)

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               'y must be a structured array with the first field'
                               ' being a binary class event indicator and the second field'
                               ' the time of the event/censoring',
                               rsvm.fit, x, y)

    def test_y_three_fields(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.ones(dtype=[('event', bool), ('time', float), ('too_much', int)], shape=10)

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               'y must be a structured array with the first field'
                               ' being a binary class event indicator and the second field'
                               ' the time of the event/censoring',
                               rsvm.fit, x, y)

    def test_event_not_boolean(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.empty(dtype=[('event', int), ('time', float)], shape=10)
        y['event'] = numpy.array([0, 1, 0, 1, 1, 0, 1, 0, 0, 1], dtype=int)
        y['time'] = numpy.arange(10)

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               "elements of event indicator must be boolean, but found int",
                               rsvm.fit, x, y)

    def test_event_not_binary(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.empty(dtype=[('event', int), ('time', float)], shape=10)
        y['event'] = numpy.array([0, 1, 2, 1, 1, 0, 1, 2, 3, 1], dtype=int)
        y['time'] = numpy.arange(10)

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               "elements of event indicator must be boolean, but found int",
                               rsvm.fit, x, y)

    def test_time_not_numeric(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.empty(dtype=[('event', bool), ('time', bool)], shape=10)
        y['event'] = numpy.array([0, 1, 0, 1, 1, 0, 1, 0, 0, 1], dtype=bool)
        y['time'] = numpy.ones(10, dtype=bool)

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               "time must be numeric, but found bool",
                               rsvm.fit, x, y)

    def test_all_censored(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=10)
        y['event'] = numpy.zeros(10, dtype=bool)
        y['time'] = numpy.array([0, 1, 2, 1, 1, 0, 1, 2, 3, 1])

        rsvm = FastSurvivalSVM()
        self.assertRaisesRegex(ValueError,
                               "all samples are censored",
                               rsvm.fit, x, y)

    def test_zero_time(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=10)
        y['event'] = numpy.array([0, 1, 0, 1, 1, 0, 1, 0, 0, 1], dtype=bool)
        y['time'] = numpy.array([0, 1, 2, 1, 1, 0, 1, 2, 3, 1])

        rsvm = FastSurvivalSVM(rank_ratio=0.5)
        self.assertRaisesRegex(ValueError,
                               "observed time contains values smaller or equal to zero",
                               rsvm.fit, x, y)

    def test_negative_time(self):
        x = numpy.arange(80).reshape(10, 8)
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=10)
        y['event'] = numpy.array([0, 1, 0, 1, 1, 0, 1, 0, 0, 1], dtype=bool)
        y['time'] = numpy.array([1, 1, -2, 1, 1, 6, 1, 2, 3, 1])

        rsvm = FastSurvivalSVM(rank_ratio=0.5)
        self.assertRaisesRegex(ValueError,
                               "observed time contains values smaller or equal to zero",
                               rsvm.fit, x, y)

    def test_ranking_with_fit_intercept(self):
        x = numpy.zeros((100, 10))
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=100)
        y['event'] = numpy.ones(100, dtype=bool)
        y['time'] = numpy.arange(1, 101, dtype=float)

        ssvm = FastSurvivalSVM(rank_ratio=1.0, fit_intercept=True)
        self.assertRaisesRegex(ValueError,
                               "fit_intercept=True is only meaningful if rank_ratio < 1.0",
                               ssvm.fit, x, y)

    def test_survial_constraints_no_ties(self):
        y = numpy.array([True, True, False, True, False, False, False, False])
        time = numpy.array([20, 46, 56, 63, 77, 90, 100, 104])

        expected_order = numpy.arange(len(time)).astype(numpy.int)

        expected = numpy.array([
            [-1, 1, 0, 0, 0, 0, 0, 0],
            [-1, 0, 1, 0, 0, 0, 0, 0],
            [-1, 0, 0, 1, 0, 0, 0, 0],
            [-1, 0, 0, 0, 1, 0, 0, 0],
            [-1, 0, 0, 0, 0, 1, 0, 0],
            [-1, 0, 0, 0, 0, 0, 1, 0],
            [-1, 0, 0, 0, 0, 0, 0, 1],
            [0, -1, 1, 0, 0, 0, 0, 0],
            [0, -1, 0, 1, 0, 0, 0, 0],
            [0, -1, 0, 0, 1, 0, 0, 0],
            [0, -1, 0, 0, 0, 1, 0, 0],
            [0, -1, 0, 0, 0, 0, 1, 0],
            [0, -1, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, -1, 1, 0, 0, 0],
            [0, 0, 0, -1, 0, 1, 0, 0],
            [0, 0, 0, -1, 0, 0, 1, 0],
            [0, 0, 0, -1, 0, 0, 0, 1],
        ], dtype=numpy.int8)

        samples_order = FastSurvivalSVM._argsort_and_resolve_ties(time, None)
        assert_array_equal(expected_order, samples_order)

        A = survival_constraints_simple(numpy.asarray(y[samples_order], dtype=numpy.uint8))
        assert_array_equal(expected, A.todense())

    def test_survival_constraints_with_ties(self):
        y = numpy.array([True, True, False, False, True, False, True, True, False, False, False, True])
        time = numpy.array([20, 33, 33, 40, 50, 66, 66, 66, 89, 110, 110, 111])

        expected_order = numpy.array([0, 2, 1, 3, 4, 7, 5, 6, 8, 9, 10, 11])
        samples_order = FastSurvivalSVM._argsort_and_resolve_ties(time, numpy.random.RandomState(0))
        numpy.testing.assert_array_equal(expected_order, samples_order)

        expected = numpy.array([
            [-1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [-1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [-1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [-1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [-1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [-1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [-1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [-1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [-1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
            [-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, -1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, -1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, -1, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, -1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, -1, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, -1, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, -1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, -1, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, -1, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, -1, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, -1, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, -1, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, -1, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, -1, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, -1, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 1],
        ], dtype=numpy.int8)

        A = survival_constraints_simple(numpy.asarray(y[samples_order], dtype=numpy.uint8))
        assert_array_equal(expected, A.todense())


@attr('slow')
class SurvivalModeCases(object):
    OPTIMIZER = None

    def setUp(self):
        x, self.y = load_whas500()
        self.x = encode_categorical(standardize(x))

    def test_default_optimizer(self):
        self.assertEqual('avltree', FastSurvivalSVM().fit(self.x.values, self.y).optimizer)

    def test_fit_and_predict_ranking(self):
        ssvm = FastSurvivalSVM(optimizer=self.OPTIMIZER, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(hasattr(ssvm, "intercept_"))
        expected_coef = numpy.array([-0.02066177, -0.26449933, -0.15205399, 0.0794547, -0.28840498, -0.02864288,
                                     0.09901995, 0.04505302, -0.12512215, 0.03341365, -0.00110442, 0.05446756,
                                     -0.53009875, -0.01394175])
        assert_array_almost_equal(expected_coef, ssvm.coef_)

        self.assertEquals(self.x.shape[1], ssvm.coef_.shape[0])

        c = ssvm.score(self.x.values, self.y)

        self.assertAlmostEqual(0.7860650174985695, c, 6)

    def test_fit_and_predict_hybrid(self):
        if self.OPTIMIZER in {'simple', 'PRSVM'}:
            raise unittest.SkipTest("regression not implemented for " + self.OPTIMIZER)

        ssvm = FastSurvivalSVM(optimizer=self.OPTIMIZER, rank_ratio=0.5,
                               max_iter=50, fit_intercept=True, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertAlmostEqual(6.1409367385513729, ssvm.intercept_)
        expected_coef = numpy.array(
            [-0.0209254120718, -0.265768317208, -0.154254689136, 0.0800600947891, -0.290121131022, -0.0288851785213,
             0.0998004550073, 0.0454100937492, -0.125863947621, 0.0343588337797, -0.000710219364914, 0.0546969104996,
             -0.5375338235, -0.0137995110308
             ])
        assert_array_almost_equal(expected_coef, ssvm.coef_)

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertAlmostEqual(780.52617631863893, rmse)

    def test_fit_and_predict_hybrid_no_intercept(self):
        if self.OPTIMIZER in {'simple', 'PRSVM'}:
            raise unittest.SkipTest("regression not implemented for " + self.OPTIMIZER)

        ssvm = FastSurvivalSVM(optimizer=self.OPTIMIZER, rank_ratio=0.5,
                               max_iter=50, fit_intercept=False, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(hasattr(ssvm, "intercept_"))
        expected_coef = numpy.array([0.00669121, -0.2754864, -0.14124808, 0.0748376, -0.2812598, 0.07543884,
                                     0.09845683, 0.08398258, -0.12182314, 0.02637739, 0.03060149, 0.11870598,
                                     -0.52688224, -0.01762842])
        assert_array_almost_equal(expected_coef, ssvm.coef_)

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertAlmostEqual(1128.4460587629746, rmse)

    def test_fit_and_predict_regression(self):
        if self.OPTIMIZER in {'simple', 'PRSVM'}:
            raise unittest.SkipTest("regression not implemented for " + self.OPTIMIZER)

        ssvm = FastSurvivalSVM(optimizer=self.OPTIMIZER, rank_ratio=0.0,
                               max_iter=50, fit_intercept=True, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertAlmostEqual(6.4160179606675278, ssvm.intercept_)
        expected_coef = numpy.array(
            [-0.0730891368237, -0.536630355029, -0.497411603275, 0.269039958377, -0.730559850692, -0.0148443526234,
             0.285916578892, 0.165960302339, -0.301749910087, 0.334855938531, 0.0886214732161, 0.0554890272028,
             -2.12680470014, 0.0421466831393
             ])
        assert_array_almost_equal(expected_coef, ssvm.coef_)

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertAlmostEqual(1206.6556186869332, rmse)

    def test_fit_and_predict_regression_no_intercept(self):
        if self.OPTIMIZER in {'simple', 'PRSVM'}:
            raise unittest.SkipTest("regression not implemented for " + self.OPTIMIZER)

        ssvm = FastSurvivalSVM(optimizer=self.OPTIMIZER, rank_ratio=0.0,
                               max_iter=50, fit_intercept=False, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(hasattr(ssvm, "intercept_"))
        expected_coef = numpy.array([1.39989875, -1.16903161, -0.40195857, -0.05848903, -0.08421557, 4.11924729,
                                     0.25135451, 1.89067276, -0.25751401, -0.10213143, 1.56333622, 3.10136873,
                                     -2.23644848, -0.11620715])
        assert_array_almost_equal(expected_coef, ssvm.coef_)

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertAlmostEqual(15838.510668936022, rmse)

    def test_fit_timeit(self):
        rnd = numpy.random.RandomState(0)
        idx = rnd.choice(numpy.arange(self.x.shape[0]), replace=False, size=100)

        ssvm = FastSurvivalSVM(optimizer=self.OPTIMIZER, timeit=3, random_state=0)
        ssvm.fit(self.x.values[idx, :], self.y[idx])

        self.assertTrue('timings' in ssvm.optimizer_result_)


class TestSimpleSurvivalMode(SurvivalModeCases, TestCase):
    OPTIMIZER = 'simple'


class TestPRSVMSurvivalMode(SurvivalModeCases, TestCase):
    OPTIMIZER = 'PRSVM'


class TestDirectCountSurvivalMode(SurvivalModeCases, TestCase):
    OPTIMIZER = 'direct-count'


class TestRBTreeSurvivalMode(SurvivalModeCases, TestCase):
    OPTIMIZER = 'rbtree'


class TestAVLTreeSurvivalMode(SurvivalModeCases, TestCase):
    OPTIMIZER = 'avltree'


class TestKernelSurvivalSVM(TestCase):

    def setUp(self):
        x, self.y = load_whas500()
        self.x = encode_categorical(standardize(x))

    def test_default_optimizer(self):
        self.assertEqual('rbtree', FastKernelSurvivalSVM().fit(self.x.values, self.y).optimizer)

    def test_unknown_optimizer(self):
        x = numpy.zeros((100, 10))
        y = numpy.empty(dtype=[('event', bool), ('time', float)], shape=100)
        y['event'] = numpy.ones(100, dtype=bool)
        y['time'] = numpy.arange(1, 101, dtype=float)

        ssvm = FastKernelSurvivalSVM(optimizer='random stuff')
        self.assertRaisesRegex(ValueError,
                               "unknown optimizer: random stuff",
                               ssvm.fit, x, y)

    def test_fit_and_predict_linear(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel='linear', random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(ssvm._pairwise)
        self.assertEquals(self.x.shape[0], ssvm.coef_.shape[0])

        i = numpy.arange(250)
        numpy.random.RandomState(0).shuffle(i)
        c = ssvm.score(self.x.values[i], self.y[i])
        self.assertAlmostEqual(0.76923445664157997, c, 6)

    def test_fit_and_predict_linear_precomputed(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel='precomputed', random_state=0)
        x = numpy.dot(self.x.values, self.x.values.T)
        ssvm.fit(x, self.y)

        self.assertTrue(ssvm._pairwise)
        self.assertEquals(self.x.shape[0], ssvm.coef_.shape[0])

        i = numpy.arange(250)
        numpy.random.RandomState(0).shuffle(i)
        c = ssvm.score(x[i], self.y[i])
        self.assertAlmostEqual(0.76923445664157997, c, 6)

    def test_fit_and_predict_linear_regression(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", rank_ratio=0.0, kernel="linear",
                                     max_iter=50, fit_intercept=True, random_state=0)

        ssvm.fit(self.x.values, self.y)

        self.assertFalse(ssvm._pairwise)
        self.assertAlmostEqual(6.3979746625712295, ssvm.intercept_, 5)

        i = numpy.arange(250)
        numpy.random.RandomState(0).shuffle(i)
        pred = ssvm.predict(self.x.values[i])
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'][i], pred))
        self.assertLessEqual(abs(1339.3006854574726 - rmse), 0.25)

    def test_fit_and_predict_linear_regression_precomputed(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", rank_ratio=0.0, kernel="precomputed",
                                     max_iter=50, fit_intercept=True, random_state=0)
        x = numpy.dot(self.x.values, self.x.values.T)
        ssvm.fit(x, self.y)

        self.assertTrue(ssvm._pairwise)
        self.assertAlmostEqual(6.3979746625712295, ssvm.intercept_, 5)

        i = numpy.arange(250)
        numpy.random.RandomState(0).shuffle(i)
        pred = ssvm.predict(x[i])
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'][i], pred))
        self.assertLessEqual(abs(1339.3006854574726 - rmse), 0.25)

    def test_fit_and_predict_linear_regression_no_intercept(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", rank_ratio=0.0, kernel="linear",
                                     max_iter=50, fit_intercept=False, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(hasattr(ssvm, "intercept_"))

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertAlmostEqual(15837.658418546907, rmse, 4)

    @attr('slow')
    def test_fit_and_predict_rbf_rbtree(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel='rbf', random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(ssvm._pairwise)
        self.assertEquals(self.x.shape[0], ssvm.coef_.shape[0])

        c = ssvm.score(self.x.values, self.y)
        self.assertAlmostEqual(0.92230102862313534, c, 3)

    @attr('slow')
    def test_fit_and_predict_rbf_avltree(self):
        ssvm = FastKernelSurvivalSVM(optimizer="avltree", kernel='rbf', random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(ssvm._pairwise)
        self.assertEquals(self.x.shape[0], ssvm.coef_.shape[0])

        c = ssvm.score(self.x.values, self.y)
        self.assertLessEqual(abs(0.92460312179802795 - c), 1e-3)

    @attr('slow')
    def test_fit_and_predict_regression_rbf(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", rank_ratio=0.0, kernel="rbf",
                                     max_iter=50, fit_intercept=True, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(ssvm._pairwise)
        self.assertAlmostEqual(4.9267218894089533, ssvm.intercept_)

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertAlmostEqual(783.525277, rmse, 6)

    @attr('slow')
    def test_fit_and_predict_hybrid_rbf(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", rank_ratio=0.5, kernel="rbf",
                                     max_iter=50, fit_intercept=True, random_state=0)
        ssvm.fit(self.x.values, self.y)

        self.assertFalse(ssvm._pairwise)
        self.assertLessEqual(abs(5.0289145697617164 - ssvm.intercept_), 0.04)

        pred = ssvm.predict(self.x.values)
        rmse = numpy.sqrt(mean_squared_error(self.y['lenfol'], pred))
        self.assertLessEqual(abs(880.20361811281487 - rmse), 75)

    @attr('slow')
    def test_fit_and_predict_clinical_kernel(self):
        x_full, y = load_whas500()

        trans = ClinicalKernelTransform()
        trans.fit(x_full)

        x = encode_categorical(standardize(x_full))

        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel=trans.pairwise_kernel, max_iter=100, random_state=0)
        ssvm.fit(x.values, y)

        self.assertFalse(ssvm._pairwise)
        self.assertEquals(x.shape[0], ssvm.coef_.shape[0])

        c = ssvm.score(x.values, y)
        self.assertLessEqual(abs(0.83699051218246412 - c), 1e-3)

    @attr('slow')
    def test_compare_rbf(self):
        x, y = load_whas500()
        x = encode_categorical(standardize(x))

        kpca = KernelPCA(kernel="rbf")
        xt = kpca.fit_transform(x)

        nrsvm = FastSurvivalSVM(optimizer='rbtree', tol=1e-8, max_iter=1000, random_state=0)
        nrsvm.fit(xt, y)

        rsvm = FastKernelSurvivalSVM(optimizer='rbtree', kernel="rbf",
                                     tol=1e-8, max_iter=1000, random_state=0)
        rsvm.fit(x, y)

        pred_nrsvm = nrsvm.predict(kpca.transform(x))
        pred_rsvm = rsvm.predict(x)

        self.assertEqual(len(pred_nrsvm), len(pred_rsvm))

        c1 = concordance_index_censored(y['fstat'], y['lenfol'], pred_nrsvm)
        c2 = concordance_index_censored(y['fstat'], y['lenfol'], pred_rsvm)

        self.assertAlmostEqual(c1[0], c2[0])
        self.assertTupleEqual(c1[1:], c2[1:])

    @attr('slow')
    def test_compare_clinical_kernel(self):
        x_full, y = load_whas500()

        trans = ClinicalKernelTransform()
        trans.fit(x_full)

        x = encode_categorical(standardize(x_full))

        kpca = KernelPCA(kernel=trans.pairwise_kernel)
        xt = kpca.fit_transform(x)

        nrsvm = FastSurvivalSVM(optimizer='rbtree', tol=1e-8, max_iter=1000, random_state=0)
        nrsvm.fit(xt, y)

        rsvm = FastKernelSurvivalSVM(optimizer='rbtree', kernel=trans.pairwise_kernel,
                                     tol=1e-8, max_iter=1000, random_state=0)
        rsvm.fit(x, y)

        pred_nrsvm = nrsvm.predict(kpca.transform(x))
        pred_rsvm = rsvm.predict(x)

        self.assertEqual(len(pred_nrsvm), len(pred_rsvm))

        c1 = concordance_index_censored(y['fstat'], y['lenfol'], pred_nrsvm)
        c2 = concordance_index_censored(y['fstat'], y['lenfol'], pred_rsvm)

        self.assertAlmostEqual(c1[0], c2[0])
        self.assertTupleEqual(c1[1:], c2[1:])

    def test_fit_precomputed_kernel_invalid_shape(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel='precomputed', random_state=0)
        x = numpy.empty((100, 14))
        y = numpy.fromiter(zip(numpy.ones(100), numpy.ones(100)), dtype=[('event', bool), ('time', float)])

        self.assertRaisesRegex(ValueError, r"Precomputed metric requires shape \(n_queries, n_indexed\)\. "
                                           r"Got \(100, 14\) for 100 indexed\.",
                               ssvm.fit, x, y)

    def test_fit_precomputed_kernel_not_symmetric(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel='precomputed', random_state=0)
        x = numpy.random.randn(100, 100)
        x[10, 12] = -1
        x[12, 10] = 9
        y = numpy.fromiter(zip(numpy.ones(100), numpy.ones(100)), dtype=[('event', bool), ('time', float)])

        self.assertRaisesRegex(ValueError, "kernel matrix is not symmetric",
                               ssvm.fit, x, y)

    def test_predict_precomputed_kernel_invalid_shape(self):
        ssvm = FastKernelSurvivalSVM(optimizer="rbtree", kernel='precomputed', random_state=0)
        x = numpy.dot(self.x.values, self.x.values.T)
        ssvm.fit(x, self.y)

        x_new = numpy.empty((100, 14))
        self.assertRaisesRegex(ValueError, r"Precomputed metric requires shape \(n_queries, n_indexed\)\. "
                                           r"Got \(100, 14\) for 500 indexed\.",
                               ssvm.predict, x_new)


class SurvivalCounterCases(object):
    TEST_CLASS = None

    def setup_01(self):
        w = numpy.array([-0.9, -0.7, -0.1, 0.15, 0.2, 1.6])
        y = numpy.array([2,       0,    4,    3,   5,   1])
        event = numpy.array([True, True, False, True, False, True])
        x = numpy.eye(6)
        v = numpy.arange(6)
        return x, y, event, w, v

    def test_calculate_01(self):
        x, y, event, w, v = self.setup_01()
        if issubclass(self.TEST_CLASS, SurvivalCounter):
            counter = self.TEST_CLASS(x, y, event, n_relevance_levels=6)
        else:
            counter = self.TEST_CLASS(x, y, event)
        counter.update_sort_order(w)

        l_plus, xv_plus, l_minus, xv_minus = counter.calculate(v)

        assert_array_equal(numpy.array([1, 4, 0, 2, 0, 4]), l_plus)
        assert_array_equal(numpy.array([2, 9, 0, 6, 0, 9]), xv_plus)
        assert_array_equal(numpy.array([2, 0, 4, 2, 3, 0]), l_minus)
        assert_array_equal(numpy.array([6, 0, 9, 6, 9, 0]), xv_minus)

    def test_calculate_01_reverse(self):
        x, y, event, w, v = self.setup_01()
        if issubclass(self.TEST_CLASS, SurvivalCounter):
            counter = self.TEST_CLASS(x, y[::-1], event[::-1], n_relevance_levels=6)
        else:
            counter = self.TEST_CLASS(x, y[::-1], event[::-1])
        counter.update_sort_order(w[::-1])

        l_plus, xv_plus, l_minus, xv_minus = counter.calculate(v[::-1])

        assert_array_equal(numpy.array([4, 0, 2, 0, 4, 1]), l_plus)
        assert_array_equal(numpy.array([9, 0, 6, 0, 9, 2]), xv_plus)
        assert_array_equal(numpy.array([0, 3, 2, 4, 0, 2]), l_minus)
        assert_array_equal(numpy.array([0, 9, 6, 9, 0, 6]), xv_minus)

    def setup_02(self):
        w = numpy.array([-0.9, -0.7, -0.1, 0.15, 0.2, 0.3, 0.8, 1.6, 1.85, 2.3])
        y = numpy.array([3,       0,    4,    6,   8,   5,   1,   7,    2,   9])
        event = numpy.array([0,   0,    0,    1,   0,   1,   1,   0,    1,   0], dtype=bool)
        x = numpy.eye(10)
        v = numpy.arange(10)
        return x, y, event, w, v

    def test_calculate_02(self):
        x, y, event, w, v = self.setup_02()
        if issubclass(self.TEST_CLASS, SurvivalCounter):
            counter = self.TEST_CLASS(x, y, event, n_relevance_levels=10)
        else:
            counter = self.TEST_CLASS(x, y, event)
        counter.update_sort_order(w)

        l_plus, xv_plus, l_minus, xv_minus = counter.calculate(v)

        assert_array_equal(numpy.array([0, 0, 0, 1, 0, 2, 6, 0, 7, 0]), l_plus)
        assert_array_equal(numpy.array([0, 0, 0, 4, 0, 7, 21, 0, 30, 0]), xv_plus)
        assert_array_equal(numpy.array([2, 0, 2, 3, 4, 2, 0, 2, 0, 1]), l_minus)
        assert_array_equal(numpy.array([14, 0, 14, 19, 22, 14, 0, 14, 0, 8]), xv_minus)

    def test_calculate_02_reverse(self):
        x, y, event, w, v = self.setup_02()
        if issubclass(self.TEST_CLASS, SurvivalCounter):
            counter = self.TEST_CLASS(x, y[::-1], event[::-1], n_relevance_levels=10)
        else:
            counter = self.TEST_CLASS(x, y[::-1], event[::-1])
        counter.update_sort_order(w[::-1])

        l_plus, xv_plus, l_minus, xv_minus = counter.calculate(v[::-1])

        assert_array_equal(numpy.array([0, 7, 0, 6, 2, 0, 1, 0, 0, 0]), l_plus)
        assert_array_equal(numpy.array([0, 30, 0, 21, 7, 0, 4, 0, 0, 0]), xv_plus)
        assert_array_equal(numpy.array([1, 0, 2, 0, 2, 4, 3, 2, 0, 2]), l_minus)
        assert_array_equal(numpy.array([8, 0, 14, 0, 14, 22, 19, 14, 0, 14]), xv_minus)


class TestSurvivalCounter(SurvivalCounterCases, TestCase):
    TEST_CLASS = SurvivalCounter


class TestRBTreeSurvivalCounter(SurvivalCounterCases, TestCase):
    TEST_CLASS = partial(OrderStatisticTreeSurvivalCounter, tree_class=RBTree)


class TestAVLTreeSurvivalCounter(SurvivalCounterCases, TestCase):
    TEST_CLASS = partial(OrderStatisticTreeSurvivalCounter, tree_class=AVLTree)


class TestNaiveSurvivalSVM(TestCase):

    def setUp(self):
        # naive survival SVM does resolve ties in survival time differently,
        # therefore use data without ties
        data = loadarff(WHAS500_NOTIES_FILE)
        x, self.y = get_x_y(data, ['fstat', 'lenfol'], '1')
        self.x = encode_categorical(x)

    def test_survival_squared_hinge_loss(self):
        nrsvm = NaiveSurvivalSVM(loss='squared_hinge', dual=False, tol=1e-8, max_iter=1000, random_state=0)
        nrsvm.fit(self.x, self.y)

        rsvm = FastSurvivalSVM(optimizer='avltree', tol=1e-8, max_iter=1000, random_state=0)
        rsvm.fit(self.x, self.y)

        assert_array_almost_equal(nrsvm.coef_.ravel(), rsvm.coef_, 3)

        pred_nrsvm = nrsvm.predict(self.x)
        pred_rsvm = rsvm.predict(self.x)

        self.assertEqual(len(pred_nrsvm), len(pred_rsvm))

        c1 = concordance_index_censored(self.y['fstat'], self.y['lenfol'], pred_nrsvm)
        c2 = concordance_index_censored(self.y['fstat'], self.y['lenfol'], pred_rsvm)

        self.assertAlmostEqual(c1[0], c2[0])
        self.assertTupleEqual(c1[1:], c2[1:])


if __name__ == '__main__':
    run_module_suite()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from .column import encode_categorical

__all__ = ['OneHotEncoder']


def check_columns_exist(actual, expected):
    missing_features = expected.difference(actual)
    if len(missing_features) != 0:
        raise ValueError("%d features are missing from data: %s" % (
            len(missing_features), missing_features.tolist()
        ))


class OneHotEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical columns with `M` categories into `M-1` columns according
    to the one-hot scheme.

    The order of non-categorical columns is preserved, encoded columns are inserted
    inplace of the original column.

    Parameters
    ----------
    allow_drop : boolean, optional, default=True
        Whether to allow dropping categorical columns that only consist
        of a single category.

    Attributes
    ----------
    `feature_names_` : pandas.Index
        List of encoded columns.

    `categories_` : dict
        Categories of encoded columns.

    `encoded_columns_` : list
        Name of columns after encoding.
        Includes names of non-categorical columns.
    """
    def __init__(self, allow_drop=True):
        self.allow_drop = allow_drop

    def fit(self, X):
        """Retrieve categorical columns.

        Parameters
        ----------
        X : pandas.DataFrame
            Data to encode.

        Returns
        -------
        self : object
            Returns self
        """
        self.fit_transform(X)
        return self

    def _encode(self, X, columns_to_encode):
        return encode_categorical(X, columns=columns_to_encode, allow_drop=self.allow_drop)

    def fit_transform(self, X, y=None, **fit_params):
        """Convert categorical columns to numeric values.

        Parameters
        ----------
        X : pandas.DataFrame
            Data to encode.
        y :
            Ignored. For computability with TransformerMixin.
        fit_params :
            Ignored. For computability with TransformerMixin.

        Returns
        -------
        Xt : pandas.DataFrame
            Encoded data.
        """
        columns_to_encode = X.select_dtypes(include=["object", "category"]).columns
        x_dummy = self._encode(X, columns_to_encode)

        self.feature_names_ = columns_to_encode
        self.categories_ = {k: X[k].cat.categories for k in columns_to_encode}
        self.encoded_columns_ = x_dummy.columns
        return x_dummy

    def transform(self, X):
        """Convert categorical columns to numeric values.

        Parameters
        ----------
        X : pandas.DataFrame
            Data to encode.

        Returns
        -------
        Xt : pandas.DataFrame
            Encoded data.
        """
        check_is_fitted(self, "encoded_columns_")
        check_columns_exist(X.columns, self.feature_names_)

        Xt = X.copy()
        for col, cat in self.categories_.items():
            Xt[col].cat.set_categories(cat, inplace=True)

        new_data = self._encode(Xt, self.feature_names_)
        return new_data.loc[:, self.encoded_columns_]



from numpy.testing import TestCase, run_module_suite, assert_array_almost_equal, assert_array_equal
import numpy

from sksurv.functions import StepFunction


class TestStepFunction(TestCase):
    def setUp(self):
        self.x = numpy.array([0, 1, 1.2, 1.75, 2, 2.1, 3, 3.94, 5.4, 9])
        self.y = numpy.array([11, 9, 9.12, 7.5, 7.25, 5.14, 3, 2.94, 2.4, 1.9])
        self.f = StepFunction(self.x, self.y)

    def test_exact(self):
        actual = numpy.array([self.f(v) for v in self.x])
        assert_array_equal(actual, self.y)

    def test_not_exact(self):
        z = numpy.diff(self.x).min() / 2
        actual = numpy.array([self.f(v + z) for v in self.x[:-1]])
        assert_array_equal(actual, self.y[:-1])

    def test_out_of_bounds(self):
        eps = numpy.finfo(numpy.float_).eps * 8
        values = [self.x[0] - 100,
                  self.x[-1] + 100,
                  self.x[0] - eps,
                  self.x[-1] + eps]

        for v in values:
            self.assertRaisesRegex(ValueError,
                                   "x must be within \[0.0+; 9.0+\], but was.+",
                                   self.f, v)

    def test_not_finite(self):
        values = [numpy.infty, -numpy.infty, numpy.nan]

        for v in values:
            self.assertRaisesRegex(ValueError,
                                   "x must be finite",
                                   self.f, v)


if __name__ == '__main__':
    run_module_suite()



from .survival_svm import FastKernelSurvivalSVM, FastSurvivalSVM
from .naive_survival_svm import NaiveSurvivalSVM
from .minlip import MinlipSurvivalAnalysis, HingeLossSurvivalSVM

__all__ = ['FastKernelSurvivalSVM',
           'FastSurvivalSVM',
           'HingeLossSurvivalSVM',
           'MinlipSurvivalAnalysis',
           'NaiveSurvivalSVM']




from .boosting import ComponentwiseGradientBoostingSurvivalAnalysis, GradientBoostingSurvivalAnalysis

__all__ = ['ComponentwiseGradientBoostingSurvivalAnalysis',
           'GradientBoostingSurvivalAnalysis']



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from sklearn.utils import check_consistent_length

__all__ = ['StepFunction']


class StepFunction(object):
    """Callable step function.

    .. math::

        f(z) = a * y_i + b,
        x_i \\leq z < x_{i + 1}

    Parameters
    ----------
    x : ndarray, shape = [n_points,]
        Values on the x axis in ascending order.

    y : ndarray, shape = [n_points,]
        Corresponding values on the y axis.

    a : float, optional
        Constant to multiply
    """
    def __init__(self, x, y, a=1., b=0.):
        check_consistent_length(x, y)
        self.x = x
        self.y = y
        self.a = a
        self.b = b

    def __call__(self, x):
        if not numpy.isfinite(x):
            raise ValueError("x must be finite")
        if x < self.x[0] or x > self.x[-1]:
            raise ValueError(
                "x must be within [%f; %f], but was %f" % (self.x[0], self.x[-1], x))
        i = numpy.searchsorted(self.x, x, side='left')
        if self.x[i] != x:
            i -= 1
        return self.a * self.y[i] + self.b

    def __repr__(self):
        return "StepFunction(x=%r, y=%r)" % (self.x, self.y)



import pandas as pd
import numpy as np
from numpy.testing import run_module_suite, TestCase
import pandas.util.testing as tm

from sksurv.preprocessing import OneHotEncoder


def create_data(n_samples=117):
    rnd = np.random.RandomState(51365192)
    data_num = pd.DataFrame(rnd.rand(n_samples, 5),
                            columns=["N%d" % i for i in range(5)])

    dat_cat = pd.DataFrame(dict(
        binary_1=pd.Categorical.from_codes(
            rnd.binomial(1, 0.6, n_samples),
            ["Yes", "No"]),
        binary_2=pd.Categorical.from_codes(
            rnd.binomial(1, 0.376, n_samples),
            ["East", "West"]),
        trinary=pd.Categorical.from_codes(
            rnd.binomial(2, 0.76, n_samples),
            ["Green", "Blue", "Red"]),
        many=pd.Categorical.from_codes(
            rnd.binomial(5, 0.47, n_samples),
            ["One", "Two", "Three", "Four", "Five", "Six"])
    ))
    data = pd.concat((data_num, dat_cat), axis=1)
    return data


def encoded_data(data):
    expected = []
    for nam, col in data.iteritems():
        if hasattr(col, "cat"):
            for cat in col.cat.categories[1:]:
                name = '{}={}'.format(nam, cat)
                s = pd.Series(col == cat, dtype=np.float64)
                expected.append((name, s))
        else:
            expected.append((nam, col))

    expected_data = pd.DataFrame.from_items(expected)
    return expected_data


class TestOneHotEncoder(TestCase):
    def test_fit(self):
        data = create_data()
        expected_data = encoded_data(data)

        t = OneHotEncoder().fit(data)

        self.assertListEqual(t.feature_names_.tolist(),
                             ['binary_1', 'binary_2', 'many', 'trinary'])
        self.assertSetEqual(set(t.encoded_columns_),
                            set(expected_data.columns))

        self.assertDictEqual(t.categories_,
                             {k: data[k].cat.categories
                              for k in ['binary_1', 'binary_2', 'many', 'trinary']})

    def test_fit_transform(self):
        data = create_data()
        expected_data = encoded_data(data)

        actual_data = OneHotEncoder().fit_transform(data)
        tm.assert_frame_equal(actual_data, expected_data)

    def test_transform(self):
        data = create_data()

        t = OneHotEncoder().fit(data)
        data = create_data(165)
        expected_data = encoded_data(data)
        actual_data = t.transform(data)
        tm.assert_frame_equal(actual_data, expected_data)

        data = pd.concat((data.iloc[:, :2], data.iloc[:, 5:], data.iloc[:, 2:5]), axis=1)
        actual_data = t.transform(data)
        tm.assert_frame_equal(actual_data, expected_data)

    def test_transform_other_columns(self):
        data = create_data()

        t = OneHotEncoder().fit(data)
        data = create_data(125)

        data_renamed = data.rename(columns={"binary_1": "renamed_1"})
        self.assertRaisesRegex(ValueError,
                               "1 features are missing from data: \['binary_1'\]",
                               t.transform, data_renamed)

        data_dropped = data.drop('trinary', axis=1)
        self.assertRaisesRegex(ValueError,
                               "1 features are missing from data: \['trinary'\]",
                               t.transform, data_dropped)

        data_renamed = data.rename(columns={"binary_1": "renamed_1", "many": "too_many"})
        self.assertRaisesRegex(ValueError,
                               "2 features are missing from data: \['binary_1', 'many'\]",
                               t.transform, data_renamed)


if __name__ == '__main__':
    run_module_suite()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy
from sklearn.utils.validation import check_consistent_length

from .util import check_y_survival

__all__ = ['kaplan_meier_estimator', 'nelson_aalen_estimator', 'ipc_weights']


def _compute_counts(event, time, order=None):
    """Count right censored and uncensored samples at each unique time point.

    Parameters
    ----------
    event : array
        Boolean event indicator.

    time : array
        Survival time or time of censoring.

    order : array or None
        Indices to order time in ascending order.
        If None, order will be computed.

    Returns
    -------
    times : array
        Unique time points.

    n_events : array
        Number of events at each time point.

    n_at_risk : array
        Number of samples that are censored or have an event at each time point.
    """
    n_samples = event.shape[0]

    if order is None:
        order = numpy.argsort(time, kind="mergesort")

    uniq_times = numpy.empty(n_samples, dtype=time.dtype)
    uniq_events = numpy.empty(n_samples, dtype=numpy.int_)
    uniq_counts = numpy.empty(n_samples, dtype=numpy.int_)

    i = 0
    prev_val = time[order[0]]
    j = 0
    while True:
        count_event = 0
        count = 0
        while i < n_samples and prev_val == time[order[i]]:
            if event[order[i]]:
                count_event += 1

            count += 1
            i += 1

        uniq_times[j] = prev_val
        uniq_events[j] = count_event
        uniq_counts[j] = count
        j += 1

        if i == n_samples:
            break

        prev_val = time[order[i]]

    times = numpy.resize(uniq_times, j)
    n_events = numpy.resize(uniq_events, j)
    total_count = numpy.resize(uniq_counts, j)

    # offset cumulative sum by one
    total_count = numpy.concatenate(([0], total_count))
    n_at_risk = n_samples - numpy.cumsum(total_count)

    return times, n_events, n_at_risk[:-1]


def _compute_counts_truncated(event, time_enter, time_exit):
    """Compute counts for left truncated and right censored survival data.

    Parameters
    ----------
    event : array
        Boolean event indicator.

    time_start : array
        Time when a subject entered the study.

    time_exit : array
        Time when a subject left the study due to an
        event or censoring.

    Returns
    -------
    times : array
        Unique time points.

    n_events : array
        Number of events at each time point.

    n_at_risk : array
        Number of samples that are censored or have an event at each time point.
    """
    if (time_enter > time_exit).any():
        raise ValueError("exit time must be larger start time for all samples")

    n_samples = event.shape[0]

    uniq_times = numpy.sort(numpy.unique(numpy.concatenate((time_enter, time_exit))), kind="mergesort")
    total_counts = numpy.empty(len(uniq_times), dtype=numpy.int_)
    event_counts = numpy.empty(len(uniq_times), dtype=numpy.int_)

    order_enter = numpy.argsort(time_enter, kind="mergesort")
    order_exit = numpy.argsort(time_exit, kind="mergesort")
    s_time_enter = time_enter[order_enter]
    s_time_exit = time_exit[order_exit]

    t0 = uniq_times[0]
    # everything larger is included
    idx_enter = numpy.searchsorted(s_time_enter, t0, side="right")
    # everything smaller is excluded
    idx_exit = numpy.searchsorted(s_time_exit, t0, side="left")

    total_counts[0] = idx_enter
    # except people die on the day they enter
    event_counts[0] = 0

    for i in range(1, len(uniq_times)):
        ti = uniq_times[i]

        while idx_enter < n_samples and s_time_enter[idx_enter] <= ti:
            idx_enter += 1

        while idx_exit < n_samples and s_time_exit[idx_exit] < ti:
            idx_exit += 1

        risk_set = numpy.setdiff1d(order_enter[:idx_enter], order_exit[:idx_exit], assume_unique=True)
        total_counts[i] = len(risk_set)

        count_event = 0
        k = idx_exit
        while k < n_samples and s_time_exit[k] == ti:
            if event[order_exit[k]]:
                count_event += 1
            k += 1
        event_counts[i] = count_event

    return uniq_times, event_counts, total_counts


def kaplan_meier_estimator(event, time_exit, time_enter=None, time_min=None):
    """Kaplan-Meier estimator of survival function.

    Parameters
    ----------
    event : array-like, shape = [n_samples,]
        Contains binary event indicators.

    time_exit : array-like, shape = [n_samples,]
        Contains event/censoring times.

    time_enter : array-like, shape = [n_samples,], optional
        Contains time when each individual entered the study for
        left truncated survival data.

    time_min : float, optional
        Compute estimator conditional on survival at least up to
        the specified time.

    Returns
    -------
    time : array, shape = [n_times]
        Unique times.

    prob_survival : array, shape = [n_times]
        Survival probability at each unique time point.
        If `time_enter` is provided, estimates are conditional probabilities.

    Example
    -------
    Creating a Kaplan-Meier curve:

    >>> x, y = kaplan_meier_estimator(event, time)
    >>> plt.step(x, y, where="post")
    >>> plt.ylim(0, 1)
    >>> plt.show()

    References
    ----------
    .. [1]: Kaplan, E. L. and Meier, P., "Nonparametric estimation from incomplete observations",
            Journal of The American Statistical Association, vol. 53, pp. 457-481, 1958.
    """
    event, time_enter, time_exit = check_y_survival(event, time_enter, time_exit)
    check_consistent_length(event, time_enter, time_exit)

    if time_enter is None:
        uniq_times, n_events, n_at_risk = _compute_counts(event, time_exit)
    else:
        uniq_times, n_events, n_at_risk = _compute_counts_truncated(event, time_enter, time_exit)

    values = 1 - n_events / n_at_risk

    if time_min is not None:
        mask = uniq_times >= time_min
        uniq_times = numpy.compress(mask, uniq_times)
        values = numpy.compress(mask, values)

    y = numpy.cumprod(values)
    return uniq_times, y


def nelson_aalen_estimator(event, time):
    """Nelson-Aalen estimator of cumulative hazard function.

    Parameters
    ----------
    y : structured array, shape = [n_samples]
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.

    Returns
    -------
    time : array, shape = [n_times]
        Unique times.

    cum_hazard : array, shape = [n_times]
        Cumulative hazard at each unique time point.

    References
    ----------
    .. [1] Nelson, W., "Theory and applications of hazard plotting for censored failure data",
           Technometrics, vol. 14, pp. 945-965, 1972.

    .. [2] Aalen, O. O., "Nonparametric inference for a family of counting processes",
           Annals of Statistics, vol. 6, pp. 701–726, 1978.
    """
    event, time = check_y_survival(event, time)
    check_consistent_length(event, time)
    uniq_times, n_events, n_at_risk = _compute_counts(event, time)

    y = numpy.cumsum(n_events / n_at_risk)

    return uniq_times, y


def ipc_weights(event, time):
    """Compute inverse probability of censoring weights

    Parameters
    ----------
    event : array, shape = [n_samples]
        Boolean event indicator.

    time_start : array, shape = [n_samples]
        Time when a subject experienced an event or was censored.

    Returns
    -------
    weights : array, shape = [n_samples]
        inverse probability of censoring weights
    """
    if event.all():
        return numpy.ones(time.shape[0])

    unique_time, p = kaplan_meier_estimator(-event, time)

    idx = numpy.searchsorted(unique_time, time[event])
    Ghat = p[idx]

    assert (Ghat > 0).all()

    weights = numpy.zeros(time.shape[0])
    weights[event] = 1.0 / Ghat

    return weights



from .stacking import Stacking
from .ensemble_selection import EnsembleSelection, EnsembleSelectionRegressor, MeanEstimator

__all__ = ['EnsembleSelection',
           'EnsembleSelectionRegressor',
           'MeanEstimator',
           'Stacking']



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numbers

import numpy

from sklearn.base import BaseEstimator
from sklearn.ensemble._gradient_boosting import _random_sample_mask
from sklearn.ensemble.base import BaseEnsemble
from sklearn.ensemble.gradient_boosting import BaseGradientBoosting, VerboseReporter
from sklearn.tree import DecisionTreeRegressor
from sklearn.tree._tree import DTYPE
from sklearn.utils import check_consistent_length, check_random_state, column_or_1d, check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import squared_norm

from scipy.sparse import csc_matrix, csr_matrix, issparse

from ..base import SurvivalAnalysisMixin
from ..util import check_arrays_survival
from .survival_loss import CoxPH, CensoredSquaredLoss, IPCWLeastSquaresError, ZeroSurvivalEstimator


__all__ = ['ComponentwiseGradientBoostingSurvivalAnalysis', 'GradientBoostingSurvivalAnalysis']


LOSS_FUNCTIONS = {"coxph": CoxPH,
                  "squared": CensoredSquaredLoss,
                  "ipcwls": IPCWLeastSquaresError}


def _sample_binomial_plus_one(p, size, random_state):
    drop_model = random_state.binomial(1, p=p, size=size)
    n_dropped = numpy.sum(drop_model)
    if n_dropped == 0:
        idx = random_state.randint(0, size)
        drop_model[idx] = 1
        n_dropped = 1
    return drop_model, n_dropped


class ComponentwiseLeastSquares(BaseEstimator):

    def __init__(self, component):
        self.component = component

    def fit(self, X, y, sample_weight):
        xw = X[:, self.component] * sample_weight
        b = numpy.dot(xw, y)
        if b == 0:
            self.coef_ = 0
        else:
            a = numpy.dot(xw, xw)
            self.coef_ = b / a

        return self

    def predict(self, X):
        return X[:, self.component] * self.coef_


def _fit_stage_componentwise(X, residuals, sample_weight, **fit_params):
    """Fit component-wise weighted least squares model"""
    n_features = X.shape[1]

    base_learners = []
    error = numpy.empty(n_features)
    for component in range(n_features):
        learner = ComponentwiseLeastSquares(component).fit(X, residuals, sample_weight)
        l_pred = learner.predict(X)
        error[component] = squared_norm(residuals - l_pred)
        base_learners.append(learner)

    # TODO: could use bottleneck.nanargmin for speed
    best_component = numpy.nanargmin(error)
    best_learner = base_learners[best_component]
    return best_learner


class ComponentwiseGradientBoostingSurvivalAnalysis(BaseEnsemble, SurvivalAnalysisMixin):
    """Gradient boosting with component-wise least squares as base learner.

    Parameters
    ----------
    loss : {'coxph', 'squared', 'ipcwls'}, optional (default='coxph')
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares error.

    learning_rate : float, optional (default=0.1)
        learning rate shrinks the contribution of each base learner by `learning_rate`.
        There is a trade-off between `learning_rate` and `n_estimators`.

    n_estimators : int (default=100)
        The number of boosting stages to perform. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.

    subsample : float, optional (default=1.0)
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. `subsample` interacts with the parameter `n_estimators`.
        Choosing `subsample < 1.0` leads to a reduction of variance
        and an increase in bias.

    dropout_rate : float, optional (default=0.0)
        If larger than zero, the residuals at each iteration are only computed
        from a random subset of base learners. The value corresponds to the
        percentage of base learners that are dropped. In each iteration,
        at least one base learner is dropped. This is an alternative regularization
        to shrinkage, i.e., setting `learning_rate < 1.0`.

    random_state : int seed, RandomState instance, or None (default=None)
        The seed of the pseudo random number generator to use when
        shuffling the data.

    verbose : int, default: 0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while (the more trees the lower the frequency). If greater
        than 1 then it prints progress and performance for every tree.

    Attributes
    ----------
    `coef_` : array, shape = [n_features]
        The aggregated coefficients.

    `loss_` : LossFunction
        The concrete ``LossFunction`` object.

    `estimators_` : list of base learners
        The collection of fitted sub-estimators.

    `train_score_` : array, shape = [n_estimators]
        The i-th score ``train_score_[i]`` is the deviance (= loss) of the
        model at iteration ``i`` on the in-bag sample.
        If ``subsample == 1`` this is the deviance on the training data.

    `oob_improvement_` : array, shape = [n_estimators]
        The improvement in loss (= deviance) on the out-of-bag samples
        relative to the previous iteration.
        ``oob_improvement_[0]`` is the improvement in
        loss of the first stage over the ``init`` estimator.

    References
    ----------
    Hothorn, T., Bühlmann, P., Dudoit, S., Molinaro, A., van der Laan, M. J.,
    "Survival ensembles", Biostatistics, 7(3), 355-73, 2006
    """
    def __init__(self, loss="coxph", learning_rate=0.1, n_estimators=100, subsample=1.0,
                 dropout_rate=0, random_state=None, verbose=0):
        self.loss = loss
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.subsample = subsample
        self.dropout_rate = dropout_rate
        self.random_state = random_state
        self.verbose = verbose

    def _check_params(self):
        """Check validity of parameters and raise ValueError if not valid. """
        if self.n_estimators <= 0:
            raise ValueError("n_estimators must be greater than 0 but "
                             "was %r" % self.n_estimators)

        if not 0.0 < self.subsample <= 1.0:
            raise ValueError("subsample must be in ]0; 1] but "
                             "was %r" % self.subsample)

        if not 0.0 < self.learning_rate <= 1.0:
            raise ValueError("learning_rate must be within ]0; 1] but "
                             "was %r" % self.learning_rate)

        if not 0.0 <= self.dropout_rate < 1.0:
            raise ValueError("dropout_rate must be within [0; 1[, but "
                             "was %r" % self.dropout_rate)

        if self.loss not in LOSS_FUNCTIONS:
            raise ValueError("Loss '{0:s}' not supported. ".format(self.loss))

    def _fit(self, X, event, time, sample_weight, random_state):
        n_samples = X.shape[0]
        # account for intercept
        Xi = numpy.column_stack((numpy.ones(n_samples), X))
        y = numpy.fromiter(zip(event, time), dtype=[('event', numpy.bool), ('time', numpy.float64)])
        y_pred = numpy.zeros(n_samples)

        do_oob = self.subsample < 1.0
        if do_oob:
            n_inbag = max(1, int(self.subsample * n_samples))

        do_dropout = self.dropout_rate > 0
        if do_dropout:
            scale = numpy.ones(int(self.n_estimators), dtype=float)

        if self.verbose:
            verbose_reporter = VerboseReporter(self.verbose)
            verbose_reporter.init(self, 0)

        for num_iter in range(int(self.n_estimators)):
            if do_oob:
                sample_mask = _random_sample_mask(n_samples, n_inbag,
                                                  random_state)
                subsample_weight = sample_weight * sample_mask.astype(numpy.float64)

                # OOB score before adding this stage
                old_oob_score = self.loss_(y[~sample_mask],
                                           y_pred[~sample_mask],
                                           sample_weight[~sample_mask])
            else:
                subsample_weight = sample_weight

            residuals = self.loss_.negative_gradient(y, y_pred, sample_weight=sample_weight)

            best_learner = _fit_stage_componentwise(Xi, residuals, subsample_weight)
            self.estimators_.append(best_learner)

            if do_dropout:
                drop_model, n_dropped = _sample_binomial_plus_one(self.dropout_rate, num_iter + 1, random_state)

                scale[num_iter] = 1. / (n_dropped + 1.)

                y_pred[:] = 0
                for m in range(num_iter + 1):
                    if drop_model[m] == 1:
                        scale[m] *= n_dropped / (n_dropped + 1.)
                    else:
                        y_pred += self.learning_rate * scale[m] * self.estimators_[m].predict(Xi)
            else:
                y_pred += self.learning_rate * best_learner.predict(Xi)

            # track deviance (= loss)
            if do_oob:
                self.train_score_[num_iter] = self.loss_(y[sample_mask], y_pred[sample_mask],
                                                         sample_weight[sample_mask])
                self.oob_improvement_[num_iter] = (old_oob_score -
                                                   self.loss_(y[~sample_mask], y_pred[~sample_mask],
                                                              sample_weight[~sample_mask]))
            else:
                # no need to fancy index w/ no subsampling
                self.train_score_[num_iter] = self.loss_(y, y_pred, sample_weight)

            if self.verbose > 0:
                verbose_reporter.update(num_iter, self)

    def fit(self, X, y, sample_weight=None):
        """
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Data matrix

        y : structured array, shape = [n_samples]
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        sample_weight : array-like, shape = [n_samples,], optional
            Weights given to each sample. If omitted, all samples have weight 1.

        Returns
        -------
        self
        """
        X, event, time = check_arrays_survival(X, y)

        n_samples, n_features = X.shape

        if sample_weight is None:
            sample_weight = numpy.ones(n_samples, dtype=numpy.float32)
        else:
            sample_weight = column_or_1d(sample_weight, warn=True)
            check_consistent_length(X, sample_weight)

        random_state = check_random_state(self.random_state)

        self._check_params()

        self.estimators_ = []
        self.n_features_ = n_features
        self.loss_ = LOSS_FUNCTIONS[self.loss](1)
        if isinstance(self.loss_, (CensoredSquaredLoss, IPCWLeastSquaresError)):
            time = numpy.log(time)

        self.train_score_ = numpy.zeros((self.n_estimators,), dtype=numpy.float64)
        # do oob?
        if self.subsample < 1.0:
            self.oob_improvement_ = numpy.zeros(self.n_estimators,
                                                dtype=numpy.float64)

        self._fit(X, event, time, sample_weight, random_state)

        return self

    def predict(self, X):
        check_is_fitted(self, 'estimators_')

        if X.shape[1] != self.n_features_:
            raise ValueError('Dimensions of X are inconsistent with training data: '
                             'expected %d features, but got %s' % (self.n_features_, X.shape[1]))

        n_samples = X.shape[0]
        Xi = numpy.column_stack((numpy.ones(n_samples), X))
        pred = numpy.zeros(n_samples, dtype=float)

        for estimator in self.estimators_:
            pred += self.learning_rate * estimator.predict(Xi)

        if isinstance(self.loss_, (CensoredSquaredLoss, IPCWLeastSquaresError)):
            numpy.exp(pred, out=pred)

        return pred

    @property
    def coef_(self):
        """Return the aggregated coefficients.

        Returns
        -------
        coef_ : array, shape = [n_features + 1]
            Coefficients of features. The first element denotes the intercept.
        """
        coef = numpy.zeros(self.n_features_ + 1, dtype=float)

        for estimator in self.estimators_:
            coef[estimator.component] += self.learning_rate * estimator.coef_

        return coef

    @property
    def feature_importances_(self):
        imp = numpy.empty(self.n_features_ + 1, dtype=object)
        for i in range(imp.shape[0]):
            imp[i] = []

        for k, estimator in enumerate(self.estimators_):
            imp[estimator.component].append(k + 1)

        def _importance(x):
            if len(x) > 0:
                return numpy.min(x)
            return numpy.nan

        ret = numpy.array([_importance(x) for x in imp])
        return ret

    def _make_estimator(self, append=True, random_state=None):
        # we don't need _make_estimator
        raise NotImplementedError()


class GradientBoostingSurvivalAnalysis(BaseGradientBoosting, SurvivalAnalysisMixin):
    """Gradient-boosted Cox proportional hazard loss with
    regression trees as base learner.

    Parameters
    ----------
    loss : {'coxph', 'squared', 'ipcwls'}, optional (default='coxph')
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares error.

    learning_rate : float, optional (default=0.1)
        learning rate shrinks the contribution of each tree by `learning_rate`.
        There is a trade-off between learning_rate and n_estimators.

    n_estimators : int (default=100)
        The number of boosting stages to perform. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.

    max_depth : integer, optional (default=3)
        maximum depth of the individual regression estimators. The maximum
        depth limits the number of nodes in the tree. Tune this parameter
        for best performance; the best value depends on the interaction
        of the input variables.
        Ignored if ``max_leaf_nodes`` is not None.

    min_samples_split : integer, optional (default=2)
        The minimum number of samples required to split an internal node.

    min_samples_leaf : integer, optional (default=1)
        The minimum number of samples required to be at a leaf node.

    min_weight_fraction_leaf : float, optional (default=0.)
        The minimum weighted fraction of the input samples required to be at a
        leaf node.

    max_features : int, float, string or None, optional (default=None)
        The number of features to consider when looking for the best split:
          - If int, then consider `max_features` features at each split.
          - If float, then `max_features` is a percentage and
            `int(max_features * n_features)` features are considered at each
            split.
          - If "auto", then `max_features=n_features`.
          - If "sqrt", then `max_features=sqrt(n_features)`.
          - If "log2", then `max_features=log2(n_features)`.
          - If None, then `max_features=n_features`.

        Choosing `max_features < n_features` leads to a reduction of variance
        and an increase in bias.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int or None, optional (default=None)
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    subsample : float, optional (default=1.0)
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. `subsample` interacts with the parameter `n_estimators`.
        Choosing `subsample < 1.0` leads to a reduction of variance
        and an increase in bias.

    dropout_rate : float, optional (default=0.0)
        If larger than zero, the residuals at each iteration are only computed
        from a random subset of base learners. The value corresponds to the
        percentage of base learners that are dropped. In each iteration,
        at least one base learner is dropped. This is an alternative regularization
        to shrinkage, i.e., setting `learning_rate < 1.0`.

    random_state : int seed, RandomState instance, or None (default=None)
        The seed of the pseudo random number generator to use when
        shuffling the data.

    verbose : int, default: 0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while (the more trees the lower the frequency). If greater
        than 1 then it prints progress and performance for every tree.


    Attributes
    ----------
    `feature_importances_` : array, shape = [n_features]
        The feature importances (the higher, the more important the feature).

    `estimators_` : ndarray of DecisionTreeRegressor, shape = [n_estimators, 1]
        The collection of fitted sub-estimators.

    `train_score_` : array, shape = [n_estimators]
        The i-th score ``train_score_[i]`` is the deviance (= loss) of the
        model at iteration ``i`` on the in-bag sample.
        If ``subsample == 1`` this is the deviance on the training data.

    `oob_improvement_` : array, shape = [n_estimators]
        The improvement in loss (= deviance) on the out-of-bag samples
        relative to the previous iteration.
        ``oob_improvement_[0]`` is the improvement in
        loss of the first stage over the ``init`` estimator.
    """
    def __init__(self, loss="coxph", learning_rate=0.1, n_estimators=100,
                 criterion='friedman_mse',
                 min_samples_split=2,
                 min_samples_leaf=1, min_weight_fraction_leaf=0.,
                 max_depth=3, min_impurity_split=1e-7, random_state=None,
                 max_features=None, max_leaf_nodes=None,
                 subsample=1.0, dropout_rate=0.0,
                 verbose=0):
        super().__init__(loss=loss,
                         learning_rate=learning_rate,
                         n_estimators=n_estimators,
                         subsample=subsample,
                         criterion=criterion,
                         min_samples_split=min_samples_split,
                         min_samples_leaf=min_samples_leaf,
                         min_weight_fraction_leaf=min_weight_fraction_leaf,
                         max_depth=max_depth,
                         min_impurity_split=min_impurity_split,
                         init=ZeroSurvivalEstimator(),
                         random_state=random_state,
                         max_features=max_features,
                         max_leaf_nodes=max_leaf_nodes,
                         verbose=verbose)
        self.dropout_rate = dropout_rate

    def _check_params(self):
        """Check validity of parameters and raise ValueError if not valid. """
        self.n_estimators = int(self.n_estimators)
        if self.n_estimators <= 0:
            raise ValueError("n_estimators must be greater than 0 but "
                             "was %r" % self.n_estimators)

        if not 0.0 < self.learning_rate <= 1.0:
            raise ValueError("learning_rate must be within ]0; 1] but "
                             "was %r" % self.learning_rate)

        if not 0.0 < self.subsample <= 1.0:
            raise ValueError("subsample must be in ]0; 1] but "
                             "was %r" % self.subsample)

        if not 0.0 <= self.dropout_rate < 1.0:
            raise ValueError("dropout_rate must be within [0; 1[, but "
                             "was %r" % self.dropout_rate)

        if isinstance(self.max_features, str):
            if self.max_features == "auto":
                max_features = self.n_features
            elif self.max_features == "sqrt":
                max_features = max(1, int(numpy.sqrt(self.n_features)))
            elif self.max_features == "log2":
                max_features = max(1, int(numpy.log2(self.n_features)))
            else:
                raise ValueError("Invalid value for max_features: %r. "
                                 "Allowed string values are 'auto', 'sqrt' "
                                 "or 'log2'." % self.max_features)
        elif self.max_features is None:
            max_features = self.n_features
        elif isinstance(self.max_features, (numbers.Integral, numpy.integer)):
            if self.max_features < 1:
                raise ValueError("max_features must be in (0, n_features]")
            max_features = self.max_features
        else:  # float
            if 0. < self.max_features <= 1.:
                max_features = max(int(self.max_features * self.n_features), 1)
            else:
                raise ValueError("max_features must be in (0, 1.0]")

        self.min_samples_split = int(self.min_samples_split)
        self.min_samples_leaf = int(self.min_samples_leaf)
        self.max_depth = int(self.max_depth)
        if self.max_leaf_nodes:
            self.max_leaf_nodes = int(self.max_leaf_nodes)

        self.max_features_ = max_features

        if self.loss not in LOSS_FUNCTIONS:
            raise ValueError("Loss '{0:s}' not supported. ".format(self.loss))

    def _fit_stage(self, i, X, y, y_pred, sample_weight, sample_mask,
                   random_state, scale, X_idx_sorted, X_csc=None, X_csr=None):
        """Fit another stage of ``n_classes_`` trees to the boosting model. """

        assert sample_mask.dtype == numpy.bool
        loss = self.loss_

        # whether to use dropout in next iteration
        do_dropout = self.dropout_rate > 0. and 0 < i < len(scale) - 1

        for k in range(loss.K):
            residual = loss.negative_gradient(y, y_pred, k=k,
                                              sample_weight=sample_weight)

            # induce regression tree on residuals
            tree = DecisionTreeRegressor(
                criterion='friedman_mse',
                splitter='best',
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                min_samples_leaf=self.min_samples_leaf,
                min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                min_impurity_split=self.min_impurity_split,
                max_features=self.max_features,
                max_leaf_nodes=self.max_leaf_nodes,
                random_state=random_state)

            if self.subsample < 1.0:
                # no inplace multiplication!
                sample_weight = sample_weight * sample_mask.astype(numpy.float64)

            if X_csc is not None:
                tree.fit(X_csc, residual, sample_weight=sample_weight,
                         check_input=False, X_idx_sorted=X_idx_sorted)
            else:
                tree.fit(X, residual, sample_weight=sample_weight,
                         check_input=False, X_idx_sorted=X_idx_sorted)

            # add tree to ensemble
            self.estimators_[i, k] = tree

            # update tree leaves
            if do_dropout:
                # select base learners to be dropped for next iteration
                drop_model, n_dropped = _sample_binomial_plus_one(self.dropout_rate, i + 1, random_state)

                # adjust scaling factor of tree that is going to be trained in next iteration
                scale[i + 1] = 1. / (n_dropped + 1.)

                y_pred[:, k] = 0
                for m in range(i + 1):
                    if drop_model[m] == 1:
                        # adjust scaling factor of dropped trees
                        scale[m] *= n_dropped / (n_dropped + 1.)
                    else:
                        # pseudoresponse of next iteration (without contribution of dropped trees)
                        y_pred[:, k] += self.learning_rate * scale[m] * self.estimators_[m, k].predict(X).ravel()
            else:
                # update tree leaves
                if X_csr is not None:
                    loss.update_terminal_regions(tree.tree_, X_csr, y, residual, y_pred,
                                                 sample_weight, sample_mask,
                                                 self.learning_rate, k=k)
                else:
                    loss.update_terminal_regions(tree.tree_, X, y, residual, y_pred,
                                                 sample_weight, sample_mask,
                                                 self.learning_rate, k=k)

        return y_pred

    def _fit_stages(self, X, y, y_pred, sample_weight, random_state,
                    begin_at_stage=0, monitor=None, X_idx_sorted=None):
        """Iteratively fits the stages.

        For each stage it computes the progress (OOB, train score)
        and delegates to ``_fit_stage``.
        Returns the number of stages fit; might differ from ``n_estimators``
        due to early stopping.
        """
        n_samples = X.shape[0]
        do_oob = self.subsample < 1.0
        sample_mask = numpy.ones((n_samples, ), dtype=numpy.bool)
        n_inbag = max(1, int(self.subsample * n_samples))
        loss_ = self.loss_

        if self.verbose:
            verbose_reporter = VerboseReporter(self.verbose)
            verbose_reporter.init(self, begin_at_stage)

        X_csc = csc_matrix(X) if issparse(X) else None
        X_csr = csr_matrix(X) if issparse(X) else None

        if self.dropout_rate > 0.:
            scale = numpy.ones(self.n_estimators, dtype=float)
        else:
            scale = None

        # perform boosting iterations
        i = begin_at_stage
        for i in range(begin_at_stage, self.n_estimators):

            # subsampling
            if do_oob:
                sample_mask = _random_sample_mask(n_samples, n_inbag,
                                                  random_state)
                # OOB score before adding this stage
                y_oob_sample = y[~sample_mask]
                old_oob_score = loss_(y_oob_sample,
                                      y_pred[~sample_mask],
                                      sample_weight[~sample_mask])

            # fit next stage of trees
            y_pred = self._fit_stage(i, X, y, y_pred, sample_weight,
                                     sample_mask, random_state, scale, X_idx_sorted,
                                     X_csc, X_csr)

            # track deviance (= loss)
            if do_oob:
                self.train_score_[i] = loss_(y[sample_mask],
                                             y_pred[sample_mask],
                                             sample_weight[sample_mask])
                self.oob_improvement_[i] = (old_oob_score - loss_(y_oob_sample, y_pred[~sample_mask],
                                                                  sample_weight[~sample_mask]))
            else:
                # no need to fancy index w/ no subsampling
                self.train_score_[i] = loss_(y, y_pred, sample_weight)

            if self.verbose > 0:
                verbose_reporter.update(i, self)

            if monitor is not None:
                early_stopping = monitor(i, self, locals())
                if early_stopping:
                    break

        if self.dropout_rate > 0.:
            self.scale_ = scale

        return i + 1

    def fit(self, X, y, sample_weight=None, monitor=None):
        """Fit the gradient boosting model.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Data matrix

        y : structured array, shape = [n_samples]
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        sample_weight : array-like, shape = [n_samples,], optional
            Weights given to each sample. If omitted, all samples have weight 1.

        monitor : callable, optional
            The monitor is called after each iteration with the current
            iteration, a reference to the estimator and the local variables of
            ``_fit_stages`` as keyword arguments ``callable(i, self,
            locals())``. If the callable returns ``True`` the fitting procedure
            is stopped. The monitor can be used for various things such as
            computing held-out estimates, early stopping, model introspect, and
            snapshoting.

        Returns
        -------
        self : object
            Returns self.
        """
        random_state = check_random_state(self.random_state)

        X, event, time = check_arrays_survival(X, y, accept_sparse=['csr', 'csc', 'coo'], dtype=DTYPE)
        n_samples, self.n_features = X.shape

        X = X.astype(DTYPE)
        if sample_weight is None:
            sample_weight = numpy.ones(n_samples, dtype=numpy.float32)
        else:
            sample_weight = column_or_1d(sample_weight, warn=True)
            check_consistent_length(X, sample_weight)

        self._check_params()

        self.loss_ = LOSS_FUNCTIONS[self.loss](1)
        if isinstance(self.loss_, (CensoredSquaredLoss, IPCWLeastSquaresError)):
            time = numpy.log(time)

        self._init_state()
        self.init_.fit(X, (event, time), sample_weight)
        y_pred = self.init_.predict(X)
        begin_at_stage = 0

        # fit the boosting stages
        y = numpy.fromiter(zip(event, time), dtype=[('event', numpy.bool), ('time', numpy.float64)])
        n_stages = self._fit_stages(X, y, y_pred, sample_weight, random_state,
                                    begin_at_stage, monitor)
        # change shape of arrays after fit (early-stopping or additional tests)
        if n_stages != self.estimators_.shape[0]:
            self.estimators_ = self.estimators_[:n_stages]
            self.train_score_ = self.train_score_[:n_stages]
            if hasattr(self, 'oob_improvement_'):
                self.oob_improvement_ = self.oob_improvement_[:n_stages]

        return self

    def _dropout_predict_stage(self, X, i, K, score):
        for k in range(K):
            tree = self.estimators_[i, k].tree_
            score += self.learning_rate * self.scale_[i] * tree.predict(X).reshape((X.shape[0], 1))
        return score

    def _dropout_decision_function(self, X):
        score = self._init_decision_function(X)

        n_estimators, K = self.estimators_.shape
        for i in range(n_estimators):
            self._dropout_predict_stage(X, i, K, score)

        return score

    def _dropout_staged_decision_function(self, X):
        X = check_array(X, dtype=DTYPE, order="C")
        score = self._init_decision_function(X)

        n_estimators, K = self.estimators_.shape
        for i in range(n_estimators):
            self._dropout_predict_stage(X, i, K, score)
            yield score.copy()

    def _scale_prediction(self, score):
        if isinstance(self.loss_, (CensoredSquaredLoss, IPCWLeastSquaresError)):
            numpy.exp(score, out=score)
        return score

    def _decision_function(self, X):
        # if dropout wasn't used during training, proceed as usual,
        # otherwise consider scaling factor of individual trees
        if not hasattr(self, "scale_"):
            score = super()._decision_function(X)
        else:
            score = self._dropout_decision_function(X)

        return self._scale_prediction(score)

    def predict(self, X):
        """Predict hazard for X.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of shape = [n_samples]
            The predicted values.
        """
        check_is_fitted(self, 'estimators_')

        X = check_array(X, dtype=DTYPE, order="C")
        score = self._decision_function(X)
        if score.shape[1] == 1:
            score = score.ravel()

        return score

    def staged_predict(self, X):
        """Predict hazard at each stage for X.

        This method allows monitoring (i.e. determine error on testing set)
        after each stage.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : generator of array of shape = [n_samples]
            The predicted value of the input samples.
        """
        check_is_fitted(self, 'estimators_')

        # if dropout wasn't used during training, proceed as usual,
        # otherwise consider scaling factor of individual trees
        if not hasattr(self, "scale_"):
            for y in self._staged_decision_function(X):
                yield self._scale_prediction(y.ravel())
        else:
            for y in self._dropout_staged_decision_function(X):
                yield self._scale_prediction(y.ravel())



from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution('scikit-survival').version
except DistributionNotFound:
    # package is not installed
    __version__ = 'unknown'



from numpy.testing import TestCase, run_module_suite

from sksurv.bintrees import RBTree, AVLTree


class BinaryTreesCases(object):
    TREE_CLASS = None

    def test_insert(self):
        tree = self.TREE_CLASS(10)
        for k in [12, 34, 45, 16, 35, 57]:
            tree.insert(k, k)
        self.assertEqual(6, len(tree))

    def test_count_smaller(self):
        tree = self.TREE_CLASS(10)
        for k in [12, 34, 45, 16, 35, 57]:
            tree.insert(k, k)

        c, a = tree.count_smaller(12)
        self.assertEqual(0, c)

        c, a = tree.count_smaller(16)
        self.assertEqual(1, c)

        c, a = tree.count_smaller(34)
        self.assertEqual(2, c)

        c, a = tree.count_smaller(35)
        self.assertEqual(3, c)

        c, a = tree.count_smaller(45)
        self.assertEqual(4, c)

        c, a = tree.count_smaller(57)
        self.assertEqual(5, c)

    def test_count_larger(self):
        tree = self.TREE_CLASS(10)
        for k in [12, 34, 45, 16, 35, 57]:
            tree.insert(k, k)

        c, a = tree.count_larger(12)
        self.assertEqual(5, c)

        c, a = tree.count_larger(16)
        self.assertEqual(4, c)

        c, a = tree.count_larger(34)
        self.assertEqual(3, c)

        c, a = tree.count_larger(35)
        self.assertEqual(2, c)

        c, a = tree.count_larger(45)
        self.assertEqual(1, c)

        c, a = tree.count_larger(57)
        self.assertEqual(0, c)


class TestRBTree(BinaryTreesCases, TestCase):
    TREE_CLASS = RBTree


class TestAVLTree(BinaryTreesCases, TestCase):
    TREE_CLASS = AVLTree


if __name__ == '__main__':
    run_module_suite()



import numpy
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import pairwise_kernels

from ..base import SurvivalAnalysisMixin
from ..util import check_arrays_survival
from ._minlip import create_difference_matrix

__all__ = ['MinlipSurvivalAnalysis', 'HingeLossSurvivalSVM']


class MinlipSurvivalAnalysis(BaseEstimator, SurvivalAnalysisMixin):
    """Survival model related to survival SVM, using a minimal Lipschitz smoothness strategy
    instead of a maximal margin strategy.

    .. math::

          \\min_{\\mathbf{w}}\\quad
          \\frac{1}{2} \\lVert \\mathbf{w} \\rVert_2^2
          + \\gamma \\sum_{i = 1}^n \\xi_i \\\\
          \\text{subject to}\\quad
          \\mathbf{w}^\\top \\mathbf{x}_i - \\mathbf{w}^\\top \\mathbf{x}_j \\geq y_i - y_j - \\xi_i,\\quad
          \\forall (i, j) \\in \\mathcal{P}_\\text{1-NN}, \\\\
          \\xi_i \\geq 0,\\quad \\forall i = 1,\\dots,n.

          \\mathcal{P}_\\text{1-NN} = \\{ (i, j) \\mid y_i > y_j \\land \\delta_j = 1
          \\land \\nexists k : y_i > y_k > y_j \\land \\delta_k = 1 \\}_{i,j=1}^n.

    Parameters
    ----------
    solver : "cvxpy" | "cvxopt", optional (default: cvxpy)
        Which quadratic program solver to use.

    alpha : float, positive  (default: 1)
        Weight of penalizing the hinge loss in the objective function.

    kernel : "linear" | "poly" | "rbf" | "sigmoid" | "cosine" | "precomputed"
        Kernel.
        Default: "linear"

    gamma : float, optional
        Kernel coefficient for rbf and poly kernels. Default: ``1/n_features``.
        Ignored by other kernels.

    degree : int (default=3)
        Degree for poly kernels. Ignored by other kernels.

    coef0 : float, optional
        Independent term in poly and sigmoid kernels.
        Ignored by other kernels.

    kernel_params : mapping of string to any, optional
        Parameters (keyword arguments) and values for kernel passed as call

    pairs : "all" | "nearest" | "next", optional (default: "nearest")
        Which constraints to use in the optimization problem.

        - all: Use all comparable pairs. Scales quadratic in number of samples
          (cf. :class:`sksurv.svm.HingeLossSurvivalSVM`).
        - nearest: Only considers comparable pairs :math:`(i, j)` where :math:`j` is the
          uncensored sample with highest survival time smaller than :math:`y_i`.
          Scales linear in number of samples.
        - next: Only compare against direct nearest neighbor according to observed time,
          disregarding its censoring status. Scales linear in number of samples.

    verbose : bool (default: False)
        Enable verbose output of solver

    timeit : False or int
        If non-zero value is provided the time it takes for optimization is measured.
        The given number of repetitions are performed. Results can be accessed from the
        ``timings_`` attribute.

    Attributes
    ----------
    `X_fit_` :
        Training data.

    `coef_` :
        Coefficients of the features in the decision function.

    References
    ----------
    .. [1] Van Belle, V., Pelckmans, K., Suykens, J. A. K., and Van Huffel, S.
           Learning transformation models for ranking and survival analysis.
           The Journal of Machine Learning Research, 12, 819-862. 2011
    """

    def __init__(self, solver="cvxpy",
                 alpha=1.0, kernel="linear", gamma=None, degree=3, coef0=1, kernel_params=None,
                 pairs="nearest", verbose=False, timeit=None):
        self.solver = solver
        self.alpha = alpha
        self.kernel = kernel
        self.gamma = gamma
        self.degree = degree
        self.coef0 = coef0
        self.kernel_params = kernel_params
        self.pairs = pairs
        self.verbose = verbose
        self.timeit = timeit

    @property
    def _pairwise(self):
        # tell sklearn.utils.metaestimators._safe_split function that we expect kernel matrix
        return self.kernel == "precomputed"

    def _get_kernel(self, X, Y=None):
        if callable(self.kernel):
            params = self.kernel_params or {}
        else:
            params = {"gamma": self.gamma,
                      "degree": self.degree,
                      "coef0": self.coef0}
        return pairwise_kernels(X, Y, metric=self.kernel,
                                filter_params=True, **params)

    def _fit(self, x, event, time):
        D = create_difference_matrix(event.astype(numpy.uint8), time, kind=self.pairs)
        K = self._get_kernel(x)

        if self.solver == "cvxpy":
            fit_func = self._fit_cvxpy
        elif self.solver == "cvxopt":
            fit_func = self._fit_cvxopt
        else:
            raise ValueError("unknown solver: {}".format(self.solver))

        if self.timeit is not None:
            import timeit

            def _inner():
                return fit_func(K, D, time)

            timer = timeit.Timer(_inner)
            self.timings_ = timer.repeat(self.timeit, number=1)

        coef, sv = fit_func(K, D, time)

        if sv is None:
            self.coef_ = coef * D
        else:
            self.coef_ = coef[:, sv] * D[sv, :]
        self.X_fit_ = x

    def _fit_cvxpy(self, K, D, time):
        import cvxpy

        n_pairs = D.shape[0]

        a = cvxpy.Variable(n_pairs)
        P = D.dot(D.dot(K).T).T
        q = D.dot(time)

        obj = cvxpy.Minimize(0.5 * cvxpy.quad_form(a, P) - a.T * q)
        assert obj.is_dcp()

        alpha = cvxpy.Parameter(sign="positive", value=self.alpha)
        constraints = [a >= 0., -alpha <= D.T * a, D.T * a <= alpha]

        prob = cvxpy.Problem(obj, constraints)
        prob.solve(verbose=self.verbose)

        return a.value.T.A, None

    def _fit_cvxopt(self, K, D, time):
        import cvxopt

        n_samples = K.shape[0]

        P = D.dot(D.dot(K).T).T
        q = -D.dot(time)

        high = numpy.repeat(self.alpha, n_samples * 2)

        n_pairs = D.shape[0]
        G = sparse.vstack((D.T, -D.T, -sparse.eye(n_pairs)))
        h = numpy.concatenate((high, numpy.zeros(n_pairs)))

        Gsp = cvxopt.matrix(G.toarray())
        # Gsp = cvxopt.spmatrix(G.data, G.row, G.col, G.shape)

        cvxopt.solvers.options["show_progress"] = int(self.verbose)
        sol = cvxopt.solvers.qp(cvxopt.matrix(P), cvxopt.matrix(q), Gsp, cvxopt.matrix(h))

        return numpy.array(sol['x']).T, None

    def fit(self, X, y):
        """Build a MINLIP survival model from training data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Data matrix.

        y : structured array, shape = [n_samples]
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        Returns
        -------
        self
        """
        X, event, time = check_arrays_survival(X, y)
        self._fit(X, event, time)

        return self

    def predict(self, X):
        """Predict risk score of experiencing an event.

        Higher scores indicate shorter survival (high risk),
        lower scores longer survival (low risk).

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of shape = [n_samples]
            Predicted risk.
        """
        K = self._get_kernel(X, self.X_fit_)
        pred = -numpy.dot(self.coef_, K.T)
        return pred.ravel()


class HingeLossSurvivalSVM(MinlipSurvivalAnalysis):
    """Naive implementation of kernel survival support vector machine.

    A new set of samples is created by building the difference between any two feature
    vectors in the original data, thus this version requires :math:`O(\\text{n_samples}^4)` space and
    :math:`O(\\text{n_samples}^6 \\cdot \\text{n_features})`.

    See :class:`sksurv.svm.NaiveSurvivalSVM` for the linear naive survival SVM based on liblinear.

    .. math::

          \\min_{\\mathbf{w}}\\quad
          \\frac{1}{2} \\lVert \\mathbf{w} \\rVert_2^2
          + \\gamma \\sum_{i = 1}^n \\xi_i \\\\
          \\text{subject to}\\quad
          \\mathbf{w}^\\top \\phi(\\mathbf{x})_i - \\mathbf{w}^\\top \\phi(\\mathbf{x})_j \\geq 1 - \\xi_{ij},\\quad
          \\forall (i, j) \\in \\mathcal{P}, \\\\
          \\xi_i \\geq 0,\\quad \\forall (i, j) \\in \\mathcal{P}.

          \\mathcal{P} = \\{ (i, j) \\mid y_i > y_j \\land \\delta_j = 1 \\}_{i,j=1,\\dots,n}.

    Parameters
    ----------
    solver : "cvxpy" | "cvxopt", optional (default: cvxpy)
        Which quadratic program solver to use.

    alpha : float, positive
        Weight of penalizing the hinge loss in the objective function (default: 1)

    kernel : "linear" | "poly" | "rbf" | "sigmoid" | "cosine" | "precomputed"
        Kernel.
        Default: "linear"

    gamma : float, optional
        Kernel coefficient for rbf and poly kernels. Default: ``1/n_features``.
        Ignored by other kernels.

    degree : int (default=3)
        Degree for poly kernels. Ignored by other kernels.

    coef0 : float, optional
        Independent term in poly and sigmoid kernels.
        Ignored by other kernels.

    kernel_params : mapping of string to any, optional
        Parameters (keyword arguments) and values for kernel passed as call

    pairs : "all" | "nearest" | "next", optional (default: "all")
        Which constraints to use in the optimization problem.

        - all: Use all comparable pairs. Scales quadratic in number of samples.
        - nearest: Only considers comparable pairs :math:`(i, j)` where :math:`j` is the
          uncensored sample with highest survival time smaller than :math:`y_i`.
          Scales linear in number of samples (cf. :class:`sksurv.svm.MinlipSurvivalSVM`).
        - next: Only compare against direct nearest neighbor according to observed time,
          disregarding its censoring status. Scales linear in number of samples.

    verbose : bool (default: False)
        Enable verbose output of solver.

    timeit : False or int
        If non-zero value is provided the time it takes for optimization is measured.
        The given number of repetitions are performed. Results can be accessed from the
        ``timings_`` attribute.

    Attributes
    ----------
    `X_fit_` :
        Training data.

    `coef_` :
        Coefficients of the features in the decision function.

    References
    ----------
    .. [1] Van Belle, V., Pelckmans, K., Suykens, J. A., & Van Huffel, S.
           Support Vector Machines for Survival Analysis. In Proc. of the 3rd Int. Conf.
           on Computational Intelligence in Medicine and Healthcare (CIMED). 1-8. 2007

    .. [2] Evers, L., Messow, C.M.,
           "Sparse kernel methods for high-dimensional survival data",
           Bioinformatics 24(14), 1632-8, 2008.

    .. [3] Van Belle, V., Pelckmans, K., Suykens, J.A., Van Huffel, S.,
           "Survival SVM: a practical scalable algorithm",
           In: Proc. of 16th European Symposium on Artificial Neural Networks,
           89-94, 2008.
    """

    def __init__(self, solver="cvxpy",
                 alpha=1.0, kernel="linear", gamma=None, degree=3, coef0=1, kernel_params=None,
                 pairs="all", verbose=False, timeit=None):
        super().__init__(solver=solver, alpha=alpha, kernel=kernel, gamma=gamma, degree=degree, coef0=coef0,
                         kernel_params=kernel_params, pairs=pairs, verbose=verbose, timeit=timeit)

    def _fit_cvxpy(self, K, D, time):
        import cvxpy

        n_pairs = D.shape[0]

        a = cvxpy.Variable(n_pairs)
        alpha = cvxpy.Parameter(sign="positive", value=self.alpha)
        P = D.dot(D.dot(K).T).T

        obj = cvxpy.Minimize(0.5 * cvxpy.quad_form(a, P) - cvxpy.sum_entries(a))
        constraints = [a >= 0., a <= alpha]

        prob = cvxpy.Problem(obj, constraints)
        prob.solve(verbose=self.verbose)

        coef = a.value.T.A
        sv = numpy.flatnonzero(coef > 1e-5)
        return coef, sv

    def _fit_cvxopt(self, K, D, time):
        import cvxopt

        n_pairs = D.shape[0]

        P = D.dot(D.dot(K).T).T
        q = -numpy.ones(n_pairs)

        G = numpy.vstack((-numpy.eye(n_pairs), numpy.eye(n_pairs)))
        h = numpy.concatenate((numpy.zeros(n_pairs), numpy.repeat(self.alpha, n_pairs)))

        cvxopt.solvers.options["show_progress"] = int(self.verbose)
        sol = cvxopt.solvers.qp(cvxopt.matrix(P), cvxopt.matrix(q), cvxopt.matrix(G), cvxopt.matrix(h))

        coef = numpy.array(sol['x']).T
        sv = numpy.flatnonzero(coef > 1e-5)
        return coef, sv



import numpy


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('ensemble', parent_package, top_path)

    config.add_extension('_coxph_loss',
                         sources=['_coxph_loss.c'],
                         include_dirs=[numpy.get_include()])

    return config


if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy

from sklearn.ensemble.gradient_boosting import LeastSquaresError, LossFunction, ZeroEstimator
from sklearn.utils.extmath import squared_norm

from ..nonparametric import ipc_weights
from ._coxph_loss import coxph_loss, coxph_negative_gradient


class ZeroSurvivalEstimator(ZeroEstimator):

    def fit(self, X, y, sample_weight=None):
        event, time = y
        return super().fit(X, time, sample_weight=sample_weight)


class CoxPH(LossFunction):
    """Cox Partial Likelihood"""

    def __call__(self, y, y_pred, sample_weight=None):
        """Compute the partial likelihood of prediction ``y_pred`` and ``y``."""
        # TODO add support for sample weights
        return coxph_loss(y['event'].astype(numpy.uint8), y['time'], y_pred.ravel())

    def negative_gradient(self, y, y_pred, sample_weight=None, **kwargs):
        """Negative gradient of partial likelihood

        Parameters
        ---------
        y : tuple, len = 2
            First element is boolean event indicator and second element survival/censoring time.
        y_pred : np.ndarray, shape=(n,):
            The predictions.
        """
        ret = coxph_negative_gradient(y['event'].astype(numpy.uint8), y['time'], y_pred.ravel())
        if sample_weight is not None:
            ret *= sample_weight
        return ret

    def init_estimator(self):
        return ZeroEstimator()

    def update_terminal_regions(self, tree, X, y, residual, y_pred,
                                sample_weight, sample_mask,
                                learning_rate=1.0, k=0):
        """Least squares does not need to update terminal regions.

        But it has to update the predictions.
        """
        # update predictions
        y_pred[:, k] += learning_rate * tree.predict(X).ravel()

    def _update_terminal_region(self, tree, terminal_regions, leaf, X, y,
                                residual, pred, sample_weight):
        pass


class CensoredSquaredLoss(LossFunction):
    """Censoring-aware squared loss.

    Censoring is taken into account by only considering the residuals
    of samples that are not censored, or the predicted survival time
    is before the time of censoring.
    """
    def __call__(self, y, y_pred, sample_weight=None):
        """Compute the partial likelihood of prediction ``y_pred`` and ``y``."""
        pred_time = y['time'] - y_pred.ravel()
        mask = (pred_time > 0) | y['event']
        return 0.5 * squared_norm(pred_time.compress(mask, axis=0))

    def negative_gradient(self, y, y_pred, sample_weight=None, **kwargs):
        """Negative gradient of partial likelihood

        Parameters
        ---------
        y : tuple, len = 2
            First element is boolean event indicator and second element survival/censoring time.
        y_pred : np.ndarray, shape=(n,):
            The predictions.
        """
        pred_time = y['time'] - y_pred.ravel()
        mask = (pred_time > 0) | y['event']
        ret = numpy.zeros(y['event'].shape[0])
        ret[mask] = pred_time.compress(mask, axis=0)
        if sample_weight is not None:
            ret *= sample_weight
        return ret

    def init_estimator(self):
        return ZeroEstimator()

    def update_terminal_regions(self, tree, X, y, residual, y_pred,
                                sample_weight, sample_mask,
                                learning_rate=1.0, k=0):
        """Least squares does not need to update terminal regions.

        But it has to update the predictions.
        """
        # update predictions
        y_pred[:, k] += learning_rate * tree.predict(X).ravel()

    def _update_terminal_region(self, tree, terminal_regions, leaf, X, y,
                                residual, pred, sample_weight):
        pass


class IPCWLeastSquaresError(LeastSquaresError):
    """Inverse probability of censoring weighted least squares error"""

    def __call__(self, y, pred, sample_weight=None):
        sample_weight = ipc_weights(y['event'], y['time'])
        return (1.0 / sample_weight.sum() *
                numpy.sum(sample_weight * ((y['time'] - pred.ravel()) ** 2.0)))

    def negative_gradient(self, y, pred, **kwargs):
        return y['time'] - pred.ravel()



from os.path import join, dirname

from nose.plugins.attrib import attr
import numpy
from numpy.testing import TestCase, run_module_suite, assert_array_almost_equal
from sklearn.model_selection import KFold, ParameterGrid
from sklearn.dummy import DummyRegressor
from sklearn.metrics import mean_squared_error

from sksurv.ensemble import ComponentwiseGradientBoostingSurvivalAnalysis
from sksurv.datasets import load_whas500
from sksurv.column import categorical_to_numeric
from sksurv.kernels import ClinicalKernelTransform
from sksurv.linear_model import IPCRidge
from sksurv.meta import EnsembleSelection, EnsembleSelectionRegressor
from sksurv.metrics import concordance_index_censored
from sksurv.svm import FastSurvivalSVM, FastKernelSurvivalSVM
from sksurv.util import check_arrays_survival




def score_cindex(est, X_test, y_test, **predict_params):
    prediction = est.predict(X_test, **predict_params)

    res = concordance_index_censored(y_test['fstat'], y_test['lenfol'], prediction)
    return res[0]


class TestEnsembleSelectionSurvivalAnalysis(TestCase):
    def setUp(self):
        x, self.y = load_whas500()
        self.x = categorical_to_numeric(x)

    def _create_ensemble(self, **kwargs):
        boosting_grid = ParameterGrid({"n_estimators": [100, 250], "subsample": [1.0, 0.75, 0.5]})
        svm_grid = ParameterGrid({"alpha": 2. ** numpy.arange(-9, 5, 2)})

        base_estimators = []
        for i, params in enumerate(boosting_grid):
            model = ComponentwiseGradientBoostingSurvivalAnalysis(random_state=0, **params)
            base_estimators.append(("gbm_%d" % i, model))

        for i, params in enumerate(svm_grid):
            model = FastSurvivalSVM(max_iter=100, random_state=0, **params)
            base_estimators.append(("svm_%d" % i, model))

        cv = KFold(n_splits=4, shuffle=True, random_state=0)
        meta = EnsembleSelection(base_estimators, n_estimators=0.4, scorer=score_cindex, cv=cv, **kwargs)
        return meta

    @attr('slow')
    def test_fit(self):
        meta = self._create_ensemble()
        self.assertEqual(len(meta), 0)

        meta.fit(self.x.values, self.y)
        self.assertEqual(len(meta), 13)
        self.assertTupleEqual(meta.scores_.shape, (13,))

        p = meta.predict(self.x.values)

        score = concordance_index_censored(self.y['fstat'], self.y['lenfol'], p)
        expected_score = numpy.array([0.7858721, 59050, 16084, 15, 119])
        assert_array_almost_equal(score, expected_score)

    @attr('slow')
    def test_fit_spearman_correlation(self):
        meta = self._create_ensemble(correlation="spearman")
        self.assertEqual(len(meta), 0)

        meta.fit(self.x.values, self.y)

        p = meta.predict(self.x.values)

        score = concordance_index_censored(self.y['fstat'], self.y['lenfol'], p)
        expected_score = numpy.array([0.7858721, 59050, 16084, 15, 119])
        assert_array_almost_equal(score, expected_score)

    @attr('slow')
    def test_fit_kendall_correlation(self):
        meta = self._create_ensemble(correlation="kendall")
        self.assertEqual(len(meta), 0)

        meta.fit(self.x.values, self.y)

        p = meta.predict(self.x.values)

        score = concordance_index_censored(self.y['fstat'], self.y['lenfol'], p)
        expected_score = numpy.array([0.7587460, 57013, 18124, 12, 119])
        assert_array_almost_equal(score, expected_score)

    @attr('slow')
    def test_fit_custom_kernel(self):
        svm_grid = ParameterGrid({"alpha": 2. ** numpy.arange(-5, 5, 2)})

        transform = ClinicalKernelTransform(fit_once=True)
        transform.prepare(self.x)

        base_estimators = []
        for i, params in enumerate(svm_grid):
            model = FastSurvivalSVM(max_iter=100, random_state=0, **params)
            base_estimators.append(("svm_linear_%d" % i, model))

        for i, params in enumerate(svm_grid):
            model = FastKernelSurvivalSVM(kernel=transform.pairwise_kernel, max_iter=100, random_state=0, **params)
            base_estimators.append(("svm_kernel_%d" % i, model))

        cv = KFold(n_splits=4, shuffle=True, random_state=0)
        meta = EnsembleSelection(base_estimators, n_estimators=0.4, scorer=score_cindex, cv=cv, n_jobs=4)

        meta.fit(self.x.values, self.y)
        self.assertEqual(len(meta), 10)
        self.assertTupleEqual(meta.scores_.shape, (10,))

        p = meta.predict(self.x.values)

        score = concordance_index_censored(self.y['fstat'], self.y['lenfol'], p)
        expected_score = numpy.array([0.7980346, 59958, 15164, 27, 119])
        assert_array_almost_equal(score, expected_score)

    def test_min_score(self):
        base_estimators = [('gbm', ComponentwiseGradientBoostingSurvivalAnalysis()),
                           ('svm', FastSurvivalSVM())]
        meta = EnsembleSelection(base_estimators, scorer=score_cindex, min_score=1.0)

        self.assertRaisesRegex(ValueError, "no base estimator exceeds min_score, try decreasing it",
                               meta.fit, self.x, self.y)

    def test_min_correlation(self):
        base_estimators = [('gbm', ComponentwiseGradientBoostingSurvivalAnalysis()),
                           ('svm', FastSurvivalSVM())]
        meta = EnsembleSelection(base_estimators, scorer=score_cindex, min_correlation=1.2)

        self.assertRaisesRegex(ValueError, "min_correlation must be in \[-1; 1\], but was 1.2",
                               meta.fit, self.x, self.y)

        meta.set_params(min_correlation=-2.1)
        self.assertRaisesRegex(ValueError, "min_correlation must be in \[-1; 1\], but was -2.1",
                               meta.fit, self.x, self.y)

        meta.set_params(min_correlation=numpy.nan)
        self.assertRaisesRegex(ValueError, "min_correlation must be in \[-1; 1\], but was nan",
                               meta.fit, self.x, self.y)

    def test_scorer(self):
        base_estimators = [('gbm', ComponentwiseGradientBoostingSurvivalAnalysis()),
                           ('svm', FastSurvivalSVM())]
        meta = EnsembleSelection(base_estimators, scorer=None)

        self.assertRaisesRegex(TypeError, "scorer is not callable",
                               meta.fit, self.x, self.y)

        meta.set_params(scorer=numpy.zeros(10))
        self.assertRaisesRegex(TypeError, "scorer is not callable",
                               meta.fit, self.x, self.y)

    def test_n_estimators(self):
        base_estimators = [('gbm', ComponentwiseGradientBoostingSurvivalAnalysis()),
                           ('svm', FastSurvivalSVM())]
        meta = EnsembleSelection(base_estimators, scorer=score_cindex, n_estimators=0)

        self.assertRaisesRegex(ValueError, "n_estimators must not be zero or negative",
                               meta.fit, self.x, self.y)

        meta.set_params(n_estimators=1000)
        self.assertRaisesRegex(ValueError, "n_estimators \(1000\) must not exceed number of base learners \(2\)",
                               meta.fit, self.x, self.y)

    def test_correlation(self):
        base_estimators = [('gbm', ComponentwiseGradientBoostingSurvivalAnalysis()),
                           ('svm', FastSurvivalSVM())]
        meta = EnsembleSelection(base_estimators, scorer=score_cindex, correlation=None)
        self.assertRaisesRegex(ValueError,
                               "correlation must be one of 'pearson', 'kendall', and 'spearman', but got None",
                               meta.fit, self.x, self.y)

        meta = EnsembleSelection(base_estimators, scorer=score_cindex, correlation=2143)
        self.assertRaisesRegex(ValueError,
                               "correlation must be one of 'pearson', 'kendall', and 'spearman', but got 2143",
                               meta.fit, self.x, self.y)

        meta = EnsembleSelection(base_estimators, scorer=score_cindex, correlation="clearly wrong")
        self.assertRaisesRegex(ValueError,
                               "correlation must be one of 'pearson', 'kendall', and 'spearman', but got 'clearly wrong'",
                               meta.fit, self.x, self.y)


def _score_rmse(est, X_test, y_test, **predict_params):
    prediction = est.predict(X_test, **predict_params)

    m = y_test['fstat']
    res = mean_squared_error(y_test['lenfol'][m], prediction[m])
    return numpy.sqrt(res)


class DummySurvivalRegressor(DummyRegressor):
    def __init__(self, strategy="mean", constant=None, quantile=None):
        super().__init__(strategy=strategy, constant=constant, quantile=quantile)

    def fit(self, X, y, sample_weight=None):
        X, event, time = check_arrays_survival(X, y)
        return super().fit(X, time)


class TestEnsembleSelectionRegressor(TestCase):
    def setUp(self):
        x, self.y = load_whas500()
        self.x = categorical_to_numeric(x)

    def _create_ensemble(self):
        aft_grid = ParameterGrid({"alpha": 2. ** numpy.arange(-9, 5, 2)})
        svm_grid = ParameterGrid({"alpha": 2. ** numpy.arange(-9, 5, 2)})

        base_estimators = []
        for i, params in enumerate(aft_grid):
            model = IPCRidge(max_iter=1000, **params)
            base_estimators.append(("aft_%d" % i, model))

        for i, params in enumerate(svm_grid):
            model = FastSurvivalSVM(rank_ratio=0, fit_intercept=True, max_iter=100,
                                    random_state=1, **params)
            base_estimators.append(("svm_%d" % i, model))

        cv = KFold(n_splits=4, shuffle=True, random_state=0)
        meta = EnsembleSelectionRegressor(base_estimators, n_estimators=0.4,
                                          scorer=_score_rmse,
                                          cv=cv, n_jobs=1)
        return meta

    @attr('slow')
    def test_fit_and_predict(self):
        meta = self._create_ensemble()
        self.assertEqual(len(meta), 0)

        meta.fit(self.x.iloc[:400].values, self.y[:400])
        self.assertEqual(len(meta), 5)
        self.assertTupleEqual(meta.scores_.shape, (14,))

        p = meta.predict(self.x.iloc[400:].values)
        score = numpy.sqrt(mean_squared_error(self.y[400:]['lenfol'], p))
        self.assertLessEqual(abs(score - 1500.01954367), 0.1)

    def test_fit_dummy(self):
        base_estimators = [
            ('dummy_0', DummySurvivalRegressor(strategy="mean")),
            ('dummy_1', DummySurvivalRegressor(strategy="median")),
            ('dummy_2', DummySurvivalRegressor(strategy="quantile", quantile=0.1)),
            ('dummy_3', DummySurvivalRegressor(strategy="quantile", quantile=0.9)),
            ('dummy_4', DummySurvivalRegressor(strategy="quantile", quantile=0.89)),
            ('dummy_5', DummySurvivalRegressor(strategy="quantile", quantile=0.91)),
        ]

        meta = EnsembleSelectionRegressor(base_estimators, n_estimators=1, min_score=5, cv=5,
                                          scorer=_score_rmse)

        self.assertRaisesRegex(ValueError, "no base estimator exceeds min_score, try decreasing it",
                               meta.fit, self.x, self.y)


if __name__ == '__main__':
    run_module_suite()



from io import TextIOWrapper, StringIO
import os
import tempfile

from numpy.testing import TestCase, run_module_suite
import numpy
import pandas
import pandas.util.testing as tm

from sksurv.io import loadarff, writearff


EXPECTED_1 = ["@relation test_nominal\n",
              "\n",
              "@attribute attr_nominal\t{beer,water,wine}\n",
              "@attribute attr_nominal_spaces\t{\"hard liquor\",mate,\"red wine\"}\n",
              "\n",
              "@data\n",
              "water,\"red wine\"\n",
              "wine,\"hard liquor\"\n",
              "beer,?\n",
              "?,mate\n",
              "wine,\"hard liquor\"\n",
              "water,mate\n"]


EXPECTED_DATETIME = ["@relation test_datetime\n",
                     "\n",
                     "@attribute attr_datetime\tdate 'yyyy-MM-dd HH:mm:ss'\n",
                     "\n",
                     "@data\n",
                     '"2014-10-31 14:13:01"\n',
                     '"2004-03-13 19:49:31"\n',
                     '"1998-12-06 09:10:11"\n']


class ReadArffTest(TestCase):
    def test_dataframe(self):
        contents = "".join(EXPECTED_1)
        with StringIO(contents) as fp:
            actual_df = loadarff(fp)

        expected_df = pandas.DataFrame.from_items(
            [("attr_nominal",
              pandas.Series(pandas.Categorical.from_codes(
                  [1, 2, 0, -1, 2, 1],
                  ["beer", "water", "wine"]))),
             ("attr_nominal_spaces",
              pandas.Series(pandas.Categorical.from_codes(
                  [2, 0, -1, 1, 0, 1],
                  ['"hard liquor"', 'mate', '"red wine"'])))
             ]
        )

        tm.assert_frame_equal(expected_df, actual_df, check_exact=True)


class WriteArffTest(TestCase):
    def test_nominal(self):
        data = pandas.DataFrame({'attr_nominal': ['water', 'wine', 'beer', None, 'wine', 'water'],
                                 'attr_nominal_spaces': ['red wine', 'hard liquor', None, 'mate', 'hard liquor',
                                                         'mate']})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                writearff(data, tfp, relation_name='test_nominal', index=False)

            with open(fp.name, 'r') as fp:
                read_date = fp.readlines()
        finally:
            os.unlink(fp.name)

        self.assertListEqual(EXPECTED_1, read_date)

    def test_nominal_with_quotes(self):
        data = pandas.DataFrame({'attr_nominal': ['water', 'wine', 'beer', None, 'wine', 'water'],
                                 'attr_nominal_spaces': ["'red wine'", "'hard liquor'", None, 'mate', "'hard liquor'",
                                                         'mate']})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                writearff(data, tfp, relation_name='test_nominal', index=False)

            with open(fp.name, 'r') as fp:
                read_date = fp.readlines()
        finally:
            os.unlink(fp.name)

        self.assertListEqual(EXPECTED_1, read_date)

    def test_nominal_as_category(self):
        data = pandas.DataFrame({'attr_nominal': pandas.Categorical(['water', 'wine', 'beer', None, 'wine', 'water'],
                                                                    ordered=False),
                                 'attr_nominal_spaces': pandas.Categorical(["'red wine'", "'hard liquor'", None,
                                                                            'mate', "'hard liquor'", 'mate'],
                                                                           ordered=False)})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                writearff(data, tfp, relation_name='test_nominal', index=False)

            with open(fp.name, 'r') as fp:
                read_date = fp.readlines()
        finally:
            os.unlink(fp.name)

        expected = EXPECTED_1.copy()
        expected[3] = "@attribute attr_nominal_spaces\t{\"hard liquor\",\"red wine\",mate}\n"
        self.assertListEqual(expected, read_date)

    def test_nominal_as_category_extra(self):
        data = pandas.DataFrame({'attr_nominal': pandas.Categorical(['water', 'wine', 'beer', None, 'wine', 'water'],
                                                                    categories=['beer', 'coke', 'water', 'wine'],
                                                                    ordered=False),
                                 'attr_nominal_spaces': pandas.Categorical(["'red wine'", "'hard liquor'", None,
                                                                            'mate', "'hard liquor'", 'mate'],
                                                                           ordered=False)})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                writearff(data, tfp, relation_name='test_nominal', index=False)

            with open(fp.name, 'r') as fp:
                read_date = fp.readlines()
        finally:
            os.unlink(fp.name)

        expected = EXPECTED_1.copy()
        expected[2] = "@attribute attr_nominal\t{beer,coke,water,wine}\n"
        expected[3] = "@attribute attr_nominal_spaces\t{\"hard liquor\",\"red wine\",mate}\n"
        self.assertListEqual(expected, read_date)

    def test_nominal_with_category_ordering(self):
        data = pandas.DataFrame({'attr_nominal': pandas.Categorical(['water', 'wine', 'beer', None, 'wine', 'water'],
                                                                    categories=['water', 'coke', 'beer', 'wine'],
                                                                    ordered=False),
                                 'attr_nominal_spaces': ["'red wine'", "'hard liquor'", None, 'mate', "'hard liquor'",
                                                         'mate']})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                writearff(data, tfp, relation_name='test_nominal', index=False)

            with open(fp.name, 'r') as fp:
                read_date = fp.readlines()
        finally:
            os.unlink(fp.name)

        expected = EXPECTED_1.copy()
        expected[2] = "@attribute attr_nominal\t{water,coke,beer,wine}\n"
        self.assertListEqual(expected, read_date)

    def test_datetime(self):
        data = pandas.DataFrame(
            {"attr_datetime": numpy.array(
                ["2014-10-31 14:13:01", "2004-03-13 19:49:31", "1998-12-06 09:10:11"], dtype="datetime64")})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                writearff(data, tfp, relation_name='test_datetime', index=False)

            with open(fp.name, 'r') as fp:
                read_date = fp.readlines()
        finally:
            os.unlink(fp.name)

        self.assertListEqual(EXPECTED_DATETIME, read_date)

    def test_unsupported_column_type(self):
        data = pandas.DataFrame(
            {"attr_datetime": numpy.array([2+3j, 45.1-1j, 0-1j, 7+0j, 132-3j, 1-0.41j], dtype="complex128")})

        fp = tempfile.NamedTemporaryFile(delete=False)
        try:
            with TextIOWrapper(fp) as tfp:
                self.assertRaisesRegex(TypeError, "unsupported type complex128",
                                       writearff, data, tfp, relation_name='test_delta', index=False)
        finally:
            os.unlink(fp.name)


if __name__ == '__main__':
    run_module_suite()
